from fireworks import factory, Message
from fireworks.utils.exceptions import EndHyperparameterOptimization
from fireworks.extensions.database import create_table
from sqlalchemy import create_engine, Column, Integer, String
from collections import defaultdict

class DummyEvaluator:

    def __init__(self): pass

    def run(self, dataloader, **kwargs): pass

def dummy_trainer():

    def trainer(parameters: dict):
        return DummyEvaluator()

    return trainer

class DummyMetric:

    def __init__(self): pass

    def attach(self, engine, name): pass

    def compute(self):
        return Message({'metric': ['hiiiii']})

def dummy_generator():

    def generator(params: list, metrics: list):
        if len(params) > 10:
            raise EndHyperparameterOptimization
        else:
            i = len(params)
            return Message({'parameters': [i]})

    return generator

def dummy_dataloader():

    return [1,2,3,4]

def test_update(): pass

def test_LocalMemoryFactory():

    trainer = dummy_trainer()
    metrics_dict = {'metric': DummyMetric()}
    generator = dummy_generator()
    dataloader = dummy_dataloader()

    memfactory = factory.LocalMemoryFactory(components={"trainer": trainer, "metrics": metrics_dict, 'parameterizer': generator, 'eval_set': dataloader})
    memfactory.run()
    params, metrics = memfactory.read()
    assert type(params) is Message
    assert type(metrics) is defaultdict
    assert set(metrics.keys()) == set(['metric'])
    assert len(params) == 11
    assert len(metrics['metric']) == 11
    assert params[5] == Message({'parameters':[5]})
    assert metrics['metric'][5] == Message({'metric': ['hiiiii']})

def test_SQLFactory():

    trainer = dummy_trainer()
    metrics_dict = {'metric': DummyMetric()}
    generator = dummy_generator()
    dataloader = dummy_dataloader()
    params_table = create_table('parameters', columns=[Column('parameters', Integer)])
    metrics_table = {'metric': create_table('metrics', columns=[Column('metric', String)])}

    engine = create_engine('sqlite:///:memory:')
    sequel = factory.SQLFactory(components={
        'trainer':trainer, 'metrics': metrics_dict, 'parameterizer': generator,
        'eval_set': dataloader, 'params_table': params_table, 'metrics_tables': metrics_table,
        'engine': engine}
        )
    sequel.run()
    params, metrics = sequel.read()

    assert type(params) is Message
    assert type(metrics) is dict
    assert set(metrics.keys()) == set(['metric'])
    assert len(params) == 11
    assert len(metrics['metric']) == 11
    assert params[5] == Message({'id':[6], 'parameters':[5]})
    assert metrics['metric'][5] == Message({'id':[6],'metric': ['hiiiii']})
    for mrow, prow in zip(metrics['metric'], params):
        assert mrow['id'][0] == prow['id'][0]



from .model import Model, model_from_module
from fireworks.utils.exceptions import ParameterizationError
from fireworks.toolbox.pipes import BatchingPipe, LoopingPipe, ShufflerPipe, RepeaterPipe
from fireworks.utils.test_helpers import DummyModel, DummyMultilinearModel, LinearJunctionModel, LinearModule, RandomJunction, generate_linear_model_data, generate_multilinear_model_data, DummyUpdateModel, ComplexModel
from fireworks import Message, Junction
import random
import torch
from torch.nn import Parameter
from random import randint
import numpy as np
import os
import shutil

loss = torch.nn.MSELoss()

def match(A, B, equivalence = lambda a,b: a is b):
    """
    Returns all elements of A that are exactly the same as an element of B. ie. a is b for some b in B.
    You can change the equivalence relation used.
    """
    matches = []
    for a in A:
        for b in B:
            if equivalence(a, b):
                matches.append(a)
                break
    return matches

def unique(A):
    """
    Returns elements of A that are unique as defined by 'is' an equivalence relation.
    """
    unique = []
    for a1 in A:
        count = 0
        for a2 in A:
            if a1 is a2:
                count +=1
                if count > 1:
                    break
        if count == 1:
            unique.append(a1)

    return unique

def train_model(model, data, models = None, predicted='y', label='y_true'):

    # Initialize model for training
    # Define loss function and learning algorithm
    models = models or [model]
    parameters = [filter(lambda p: p.requires_grad, m.parameters()) for m in models]
    parameters = [x for y in parameters for x in y]
    optimizer = torch.optim.SGD(parameters, lr=.00015)
    # Training loop
    num_epochs = 1
    for epoch in range(num_epochs):
        for batch in data:
            optimizer.zero_grad()
            result = model(batch.to_tensors())
            lo = loss(result[predicted], result[label])
            lo.backward()
            optimizer.step()

    return model

def test_Model_init():

    # Confirm that the model detects that parameters are missing.
    # try:
    #     damura = DummyModel()
    #     assert False
    # except ParameterizationError as error:
    #     assert True

    # Confirm that the model can be initialized
    # damura = DummyModel({'m': [3.]})
    #
    # assert (damura.m == 3.).all()
    # assert (damura.b == 0).all() # Default for 'b'

    damura = DummyModel({'m': [4.], 'b': [5.]})
    assert (damura.m == 4.).all()
    assert (damura.b == 5.).all()

def test_Model_get_set_state():
    sabura = DummyModel({'m': [2.], 'b': [3.]})
    mabura = DummyModel({'m': (sabura, 'm')})
    state1 = sabura.get_state()
    assert state1['external'] == {}
    assert (state1['internal'].keys()) == set(['out_column', 'in_column', 'b', 'm'])
    assert state1['internal']['b'] == 3.
    state2 = mabura.get_state()
    assert state2['external']['m'] == (sabura, 'm')
    assert (state2['internal'].keys()) == set(['out_column', 'in_column', 'b'])
    assert state2['internal']['b'] == 0.
    cabura = DummyModel()
    cabura.set_state(state1)
    cstate1 = cabura.get_state()
    assert cstate1['external'] == {}
    assert (cstate1['internal'].keys()) == set(['out_column', 'in_column', 'b', 'm'])
    assert cstate1['internal']['b'] == 3.
    cabura.set_state(state2)
    cstate2 = cabura.get_state()
    assert cstate2['external']['m'] == (sabura, 'm')
    assert (cstate2['internal'].keys()) == set(['out_column', 'in_column', 'b'])
    assert cstate2['internal']['b'] == 0.
    neuralnet = ComplexModel()
    state = neuralnet.get_state()
    neuralnet2 = ComplexModel()
    neuralnet2.set_state(state, reset=False)

def test_Model_save_load():

    # Construct model with inputs and components.
    sabura = DummyModel({'m': [0.]})
    sabura.__name__ = 'sabura'
    damura = DummyModel({'m': [2.]})
    damura.__name__ = 'damura'
    lamura = DummyModel()
    lamura.__name__ = 'lamura'
    babura = DummyModel({'m': (damura, 'm'), 'b': [4.], 'c': sabura, 'd': torch.nn.Conv2d(4,5,4)}, input=lamura)
    babura.__name__ = 'babura'

    if not os.path.isdir('save_test'):
        os.mkdir('save_test')
    else:
        shutil.rmtree('save_test')
        os.mkdir('save_test')
    babura.save(method='json', path='save_test/test.json')
    files = os.walk('save_test/.').__next__()[2]
    assert len(files) == 3
    # Test different save methods.
    newone = DummyModel({'m': [5.], 'd': torch.nn.Conv2d(4,5,4), 'c': DummyModel()})
    assert (newone.m == 5.).all()
    # old = newone.state_dict()['d.bias'].clone().detach().numpy()
    newone.load_state('save_test/torch_babura-test.json')
    assert (newone.d.bias - babura.d.bias < .005).all()
    assert (newone.d.weight - babura.d.weight < .005).all()
    # new = newone.state_dict()['d.bias'].clone().detach().numpy()
    shutil.rmtree('save_test')

def test_Model_inferencing():

    damura = DummyModel({'m': [2.]})
    x = Message({'x': torch.Tensor([1,2,3])})
    y = damura(x)
    assert y == x
    assert (y['x'] == torch.Tensor([1,2,3])).all()
    assert (y['y'] == torch.Tensor([2.,4.,6.])).all()

def test_ModelFromModule():

    pop = LinearModule()
    mom = model_from_module(LinearModule)
    messi = Message({'x':torch.Tensor([1,2,3])})
    pom = mom()
    assert set(pom.components) == set(['conv1', 'm', 'b'])
    result = pom(messi)
    y = result['y']
    x = result['x']
    assert 'y' in result and 'x' in result and 'z' not in result
    # Test that Model calls are recursive on inputs
    hom = DummyModel({'m': [3.]}, out_column='z')
    pom.input = hom
    result = pom(messi)
    assert (x == result['x']).all()
    assert (y == result['y']).all()
    assert 'y' in result and 'x' in result and 'z' in result

def test_save_model():

    A = DummyModel({'m': [0.], 'ok': torch.nn.Conv2d(1,20,5), 'c': 'yes'})
    A.save(method='json')
    A.save(method='html')
    saved = A.save(method='dict')
    assert type(saved) is dict
    assert set(saved.keys()) == set(['out_column', 'm', 'in_column', 'ok', 'c', 'b'])

def test_freeze_and_unfreeze():

    A = DummyModel({'m': [0.]})
    assert A.m.requires_grad == True
    assert A.b.requires_grad == True
    A.freeze()
    assert A.m.requires_grad == False
    assert A.b.requires_grad == False
    A.unfreeze()
    assert A.m.requires_grad == True
    assert A.b.requires_grad == True
    A.freeze('m')
    assert A.m.requires_grad == False
    assert A.b.requires_grad == True
    A.unfreeze('m')
    assert A.m.requires_grad == True
    assert A.b.requires_grad == True
    A.freeze('b')
    assert A.m.requires_grad == True
    assert A.b.requires_grad == False
    A.unfreeze('b')
    assert A.m.requires_grad == True
    assert A.b.requires_grad == True

    B = model_from_module(LinearModule)() # This one has a PyTorch module Conv1 in it as well.
    assert B.m.requires_grad == True
    assert B.b.requires_grad == True
    assert B.conv1.weight.requires_grad == True
    assert B.conv1.weight.requires_grad == True
    B.freeze()
    assert B.m.requires_grad == False
    assert B.b.requires_grad == False
    assert B.conv1.weight.requires_grad == False
    assert B.conv1.weight.requires_grad == False
    B.unfreeze('conv1')
    assert B.m.requires_grad == False
    assert B.b.requires_grad == False
    assert B.conv1.weight.requires_grad == True
    assert B.conv1.weight.requires_grad == True
    B.freeze('conv1')
    assert B.m.requires_grad == False
    assert B.b.requires_grad == False
    assert B.conv1.weight.requires_grad == False
    assert B.conv1.weight.requires_grad == False
    B.unfreeze(['m','b'])
    assert B.m.requires_grad == True
    assert B.b.requires_grad == True
    assert B.conv1.weight.requires_grad == False
    assert B.conv1.weight.requires_grad == False

def get_minibatcher(training_data):

    repeater = RepeaterPipe(training_data)
    lol = LoopingPipe(repeater)
    shuffler = ShufflerPipe(lol)
    minibatcher = BatchingPipe(shuffler, batch_size=50)

    return minibatcher

def test_one_Model_training():

    A = DummyModel({'m': [0.]})
    B = model_from_module(LinearModule)()
    training_data = generate_linear_model_data()
    m = training_data[1]['m']
    b = training_data[1]['b']
    errors = training_data[1]['errors']
    minibatcher = get_minibatcher(training_data[0])
    train_model(A, minibatcher)
    # For some reason, this model struggles to learn the y-intercept.
    assert (m-A.m < .6).all()
    train_model(B, minibatcher)
    assert (m - B.m < .6).all()

    assert (A.m - B.m < .6).all() # Test precision between models

def test_multiple_Models_training():
    """
    Here, we compose a model multilinear = A + B + C, where A, B, and C represent
    different components of the overall model.
    """
    A = DummyModel({'m': [0.]})
    A.freeze('b')
    B = model_from_module(LinearModule)()
    B.freeze('b')
    C = DummyModel({'m': [0.]})
    C.freeze('m')
    multilinear = DummyMultilinearModel({'m1':A.m, 'm2': B.m, 'b': C.b})
    assert multilinear.m1 is A.m
    assert multilinear.m2 is B.m
    assert multilinear.b is C.b
    training_data = generate_multilinear_model_data()
    m1 = training_data[1]['m1']
    m2 = training_data[1]['m2']
    b = training_data[1]['b']
    errors = training_data[1]['errors']
    minibatcher = get_minibatcher(training_data[0])
    train_model(multilinear, minibatcher)
    assert (A.m - m1 < .6).all()
    assert (B.m - m2 < .6).all()
    assert (A.b == 0).all()
    assert (C.m == 0.).all()

def test_setattr_in_pipeline():
    """
    Setattribute should not trigger any recursive actions.
    """
    A = DummyModel({'m': [3.]}, out_column='y1')
    B = DummyModel({'m': [1.], 'b': [2.]}, input=A, in_column='y1', out_column='y')
    A.ok = 0
    B.ok = 1
    assert A.ok == 0
    assert B.ok == 1
    assert (A.m == 3.).all()
    assert (B.m == 1.).all()
    # Test that Parameters and subModules automatically get added as components.
    A.yes = Parameter(torch.Tensor([3.,4.]))
    assert 'yes' in A.components
    assert 'yes' not in B.components
    assert B.yes is A.yes

def test_all_parameters():

    # Pipe layout
    A = DummyModel({'m': [3.]}, out_column='y1')
    B = DummyModel({'m': [1.], 'b': [2.]}, input=A, in_column='y1', out_column='y')
    # assert len(unique(B.all_parameters())) == len(B.all_parameters())
    # matches = match(B.all_parameters(), [B.m, B.b, A.m, A.b])
    # assert len(matches) == 4

    # Component layout
    A = DummyModel({'m': [0.]})
    B = model_from_module(LinearModule)()
    C = DummyModel({'m': [1.]})
    multilinear = DummyMultilinearModel({'m1':(A, 'm'), 'm2': (B, 'm'), 'b': (C, 'm'), 'ok': [1.]})
    multilinear.state_dict()
    multilinear.hi = torch.nn.Conv2d(4,4,4)
    params = multilinear.all_parameters()
    assert len(params) == 6

    for param in params: # See if teh Conv2d from above had it's parameters registered.
        if param.shape == (4,4,4,4,):
            test = True
    assert test


    multilinear2 = DummyMultilinearModel()
    multilinear2.hi = torch.nn.Conv2d(4,4,4)
    state = multilinear.get_state()
    multilinear2.set_state(state, reset=False)
    matches = match(multilinear.all_parameters(), multilinear2.all_parameters(), equivalence = lambda a,b: (a==b).all())
    assert len(matches) == 6

    assert len(unique(multilinear.all_parameters())) == len(multilinear.all_parameters())
    matches = match(multilinear.all_parameters(), [A.m, B.m, C.m])
    assert len(matches) == 3

    # Component layout
    A = DummyModel({'m': [1.],'b':[3.]}, out_column='z')
    B = LinearJunctionModel(components={'b':[.5], 'f':A})
    assert len(unique(B.all_parameters())) == len(B.all_parameters())
    matches = match(B.all_parameters(), [A.m, A.b, B.b])
    assert len(matches) == 3

    # Pipes and components
    A = DummyModel({'m': [1.],'b':[3.]}, out_column='z')
    C = DummyModel({'m': [0.]})
    B = LinearJunctionModel(components={'b':[.5], 'f':A}, input=C)

    assert len(unique(B.all_parameters())) == len(B.all_parameters())
    matches = match(B.all_parameters(), [A.m, A.b, B.b, C.m, C.b])
    assert len(matches) == 5

def test_multiple_Models_training_in_pipeline():
    """
    Here, model A pipes its output into B
    """
    A = DummyModel({'m': [3.]}, out_column='y1')
    B = DummyModel({'m': [1.], 'b': [2.]}, input=A, in_column='y1', out_column='y')
    A.freeze('b')
    B.freeze('m')
    training_data = generate_linear_model_data()
    m = training_data[1]['m']
    b = training_data[1]['b']
    errors = training_data[1]['errors']
    minibatcher = get_minibatcher(training_data[0])
    assert (A.m == 3.).all()
    assert (B.m == 1).all()
    assert (A.b == 0).all()
    assert (B.b == 2.).all()
    batch = minibatcher.__next__()
    batch.to_tensors()
    A(batch)
    train_model(B, minibatcher, models = [A, B])
    assert (A.m - m < .6).all()
    assert (B.b != 2).all()
    assert (B.m == 1).all()
    assert (A.b == 0).all()

def test_multiple_Models_training_in_junction():
    """
    Here, model A is provided as a component of B
    """
    A = DummyModel({'m': [1.],'b':[3.]}, out_column='z')
    B = LinearJunctionModel(components={'b':[.5], 'f':A})
    training_data = generate_linear_model_data()
    m = training_data[1]['m']
    b = training_data[1]['b']
    errors = training_data[1]['errors']
    minibatcher = get_minibatcher(training_data[0])
    batch = minibatcher.__next__()
    batch.to_tensors()
    B(batch)
    assert (A.m == 1.).all()
    train_model(B, minibatcher, models=[A, B])
    assert (A.m != 1).all()

def test_multiple_Models_training_via_junction():
    """
    Here, model B takes a junction A as input that randomly calls one of C, D, or E
    Hence, B implements y = f(x) + b, where f is randomly C, D, or E
    """
    C = DummyModel({'m': [1.],'b':[0.]}, out_column='z')
    D = DummyModel({'m': [2.],'b':[0.]}, out_column='z')
    E = DummyModel({'m': [3.],'b':[0.]}, out_column='z')
    C.freeze('b')
    D.freeze('b')
    E.freeze('b')
    A = RandomJunction(components={'C':C, 'D': D, 'E':E})
    B = LinearJunctionModel(components={'b': [3.], 'f':A})
    training_data = generate_linear_model_data(n=750)
    m = training_data[1]['m']
    b = training_data[1]['b']
    errors = training_data[1]['errors']
    minibatcher = get_minibatcher(training_data[0])
    batch = minibatcher.__next__()
    batch.to_tensors()
    banana = B(batch)['y']
    rambo = False
    for i in range(20):
        bonana = B(batch)['y']
        if (banana != bonana).all():
            rambo = True
            break
    assert rambo
    train_model(B, minibatcher, models=[B, C, D, E])
    # Test that all Models trained
    assert (C.m - m < .6).all()
    assert (D.m - m < .6).all()
    assert (E.m - m < .6).all()
    assert (C.m != D.m).all()
    assert (D.m != E.m).all()
    assert (E.m != C.m).all()

def test_enable_disable():

    def get_dummys():
        """
        This Model increments it's internal _count variable by the length of the Message being accessed on update.
        On inference, it simply appends the input to itself and returns that.
        """
        dummy = DummyUpdateModel()
        tummy = DummyUpdateModel(input=dummy)
        assert dummy._count == 0
        assert tummy._count == 0
        return dummy, tummy

    data, meta = generate_linear_model_data()
    batch = data[2:10]
    l = len(batch)

    # Test updates and inference for one model
    dummy, tummy = get_dummys()
    out = dummy(batch)
    assert dummy._count == l
    assert tummy._count == 0
    assert out == batch.append(batch)

    # Test updates and inference for multiple Models
    dummy, tummy = get_dummys()
    out = tummy(batch)
    assert dummy._count == l
    assert tummy._count == 2*l
    assert out == batch.append(batch).append(batch).append(batch) # 4x

    # Test disable updates for one Model
    dummy, tummy = get_dummys()
    tummy.disable_updates()
    out = tummy(batch)
    assert dummy._count == l
    assert tummy._count == 0
    assert out == batch.append(batch).append(batch).append(batch) # 4x

    # Test disable updates for all Models recursively
    dummy, tummy = get_dummys()
    tummy.disable_updates_all()
    out = tummy(batch)
    assert dummy._count == 0
    assert tummy._count == 0
    assert out == batch.append(batch).append(batch).append(batch) # 4x
    # Test reenabling updates for one Model
    tummy.enable_updates()
    out = tummy(batch)
    assert dummy._count == 0
    assert tummy._count == 2*l
    assert out == batch.append(batch).append(batch).append(batch) # 4x
    # Test reenabling updates for all Models
    tummy.enable_updates_all()
    out = tummy(batch)
    assert dummy._count == l
    assert tummy._count == 4*l
    assert out == batch.append(batch).append(batch).append(batch) # 4x

    # Test disable inference for one Model
    dummy, tummy = get_dummys()
    tummy.disable_inference()
    out = tummy(batch)
    assert dummy._count == l
    assert tummy._count == 2*l
    assert out == batch.append(batch) # 2x from the first one

    # Test disable inference for all Models recursively
    dummy, tummy = get_dummys()
    tummy.disable_inference_all()
    out = tummy(batch)
    assert dummy._count == l
    assert tummy._count == l
    assert out == batch # Identity
    # Test renabling inference for one Model
    tummy.enable_inference()
    out = tummy(batch)
    assert dummy._count == 2*l
    assert tummy._count == 2*l
    assert out == batch.append(batch) # 2x from the first one
    # Test renabling inference for all Models
    tummy.enable_inference_all()
    out = tummy(batch)
    assert dummy._count == 3*l
    assert tummy._count == 4*l
    assert out == batch.append(batch).append(batch).append(batch) # 4x from the first one



import torch
from .message import Message
from torch.nn import Parameter, Module

class Component_Map(dict):
    """
    Each of the main objects that can be used to construct pipelines (Pipes, Junctions, and Models) have a means for tracking their internal
    state. For Pipes, this is handled by a simple dictionary, but for Junctions and Models we use this class which satisfies the more complex
    needs for these objects.
    In particular, a Component_Map can track whether a variable is internal or external to a given object, and if it's external, whom the
    variable belongs to. This lets us dynamically assign a variable of one object to to a component of a Junction or Model while maintaining
    the distinction that the assigned variable is not internal to Junction or Model.
    This distinction can be useful for variables such as hyperparameters or runtime configurations (eg. whether to use Cuda) that one does
    not want to store alongside variables like model weights. You can also have a Model 'borrow' variables from another Model while maintaining
    this distinciton (eg. use the first two layers from this other model, then use the remainder of the layers using internal weights), and
    this can be useful when training Models (you could have your optimizer operate only on a Model's internal parameters, treating everything
    else as constant.)
    These are a just few examples of how this abstraction can be useful, and in simpler terms, it is essentially a means to deliberately pass
    variables by reference, which is not how Python's memory model operates by default, but it can be extremely helpful when doing machine
    learning.
    The details of the interaction with a Component_Map are abstracted away by Junctions and Models. Hence, you shouldn't have to directly
    interact with a Component_Map. Instead, you can generally just call set_state and get_state on Junctions and Models to get serialized
    representations of the Component_Maps.

    The format of these serialization is a dict of the form {'external': {...}, 'internal': {...}}.
    The 'internal' dict contains a mapping between variable names and those variables.
    The 'external' dict contains a mapping between variable names and the object that those variables belong to. In this way, a Component_Map
    can keep track of the owner of the linked variable and also get its value as needed. Hence, Junctions and Models can simply use that
    variable as if it were internal, and this makes it easy to swap variables around without changing syntax (eg. replace some internal
    component of a Model with an attribute of some other object on the fly.)

    A Component_Map behaves like a dict with the special property that if you assign an tuple of the form (obj, x) to the dict, where
    x is a string, then the Component_Map will treat that as a 'pass by reference' assignment. In other words, it will assume that you
    want to externally link the variable obj.x to the Component_Map. For example, if you do this:

    ::

        A = some_object()
        cm = Component_Map()
        cm['a'] = (A, 'x')

    Now whenever you call cm['a'], you will get whatever is returned by A.x.

    ::

        cm['a'] == A.x # This evaluates to True.
        cm['a'] is A.x # This also evaluates to True, because the assignment is by reference.

    If you cm.get_state(), the 'external' dict will contain a reference to A.

    ::

        state = cm.get_state()
        external = state['external']
        external['a'] == (A, 'x') # This evaluates to True.

    On the other hand, if you do this:

    ::

        cm['a'] = A.x # Don't pass by reference.
        cm['a'] == A.x # This evaluates to True.
        cm['a'] is A.x # This may or may not be True because Python sometimes assigns by reference and sometimes copies data depending on the situation.

    This will be treated as an internal assignment. Note that PyTorch implements logic for enforcing pass-by-reference for torch.nn.Parameter
    objects. Hence, if A.x was a Parameter, then the assignment will be by reference. However, we will have no way of knowing who the
    'owner' of the Parameter is, and by using Component_Maps, we also are able to extend this functionality to any Python object. If
    you now get the state, it will be in the 'internal' dict.

    ::

        state = cm.get_state()
        internal = state['internal']
        internal['a'] == A.x # This evaluates to True. If A.x is vector/tensor-valued, you may get a vector/tensor of 1's.

    """

    def __init__(self, components):
        """
        self._internal_components functions like a normal dictionary. If you call cm['a'] and 'a' is in the internal_components dict, then
        it will return cm._internal_components['a']. However, if 'a' is external, then cm['a'] will return
        getattr(cm._external_modules['a'], c._external_attribute_names['a']).
        """
        dict.__init__(self)
        self._external_modules = {}
        self._external_attribute_names = {}
        self._internal_components = {}

        for key, value in components.items():
            self[key] = value

    def __setitem__(self, key, val):
        """
        This overrides the __setitem__ method of dict so that if you set an item of the form (obj, x) to a key k, where x is a string,
        the Component_Map will 'link' obj.x to k. ie. cm['k'] == obj.x. It will do this by inserting obj into the _external_modules dict
        and x into the _external_attribute_names dict with k as the key for both. Hence, when you call cm['k'], the Component_Map will fetch
        obj and the attribute name x and then call getattr(obj, x).

        Note that you cannot have a key in both the internal and external dicts. If you already have a key in one and you assign to the other,
        the former will be deleted. This prevents ambiguity when accessing elements.
        """
        if type(val) is tuple and len(val) is 2 and type(val[1]) is str and hasattr(val[0], val[1]):
            # Very specific test to check if the intention is to link an attribute inside of another
            # object to this Component Map rather than simply set the value of the key to a tuple.

            if key in self._internal_components:
                # Delete from internal components if this key already exists
                del self[key]
            obj, attribute = val
            value = getattr(obj, attribute)
            # key, value = self.setitem_hook(key, value)
            # self._external_components[key] = value
            self._external_modules[key] = obj
            self._external_attribute_names[key] = attribute
        else:
            if key in self._external_modules:
                # Deelte from external components if this key already exists.
                del self[key]
            value = val
            key, value = self.setitem_hook(key, value)
            self._internal_components[key] = value

        dict.__setitem__(self, key, value)

    def __delitem__(self, key):
        """
        This also deletes the key from the the internal and external dicts.
        """

        try:
            dict.__delitem__(self, key)
            for d in [self._external_attribute_names, self._external_modules, self._internal_components]:
                if key in d:
                    del d[key]
        except KeyError:
            raise KeyError

    def __getitem__(self, key):
        """
        This returns the object referenced by the key, whether it is internal or external.
        """
        if key in self._internal_components:
            return self._internal_components[key]
        elif key in self._external_modules:
            return getattr(self._external_modules[key], self._external_attribute_names[key])
        else:
            raise AttributeError()

    def setitem_hook(self, key, value):
        """
        This can be overridden by a subclass in order to implement specific actions
        that should take place before an attribute is set.
        """
        return key, value

    def set_state(self, state):
        """
        This method can be used to apply a serialized representation of state to a Component_Map at once. This is used for loading in saved
        data.

        Args:
            state: A dict of the form {'external': {...}, 'internal': {...}}. The elements of this dict will be assigned to the Component_Map.
                   Note that this will not reset the Component_Map, so if there were previous elements already present, those will remain
                   in the Component_Map.
        """
        for key, value in {**state['internal'], **state['external']}.items():

            self[key] = value

    def get_state(self):
        """
        This returns a serialized representation of the state of the Component_Map.

        Args:
            None

        Returns:
            state: A dict of the form {'external': {...}, 'internal': {...}}. See above documentation for more information.
        """
        internal = self._internal_components
        external = {key: (self._external_modules[key], self._external_attribute_names[key]) for key in self._external_modules}

        return {'internal': internal, 'external': external}

class PyTorch_Component_Map(Component_Map):
    """
    This is a subclass of Component_Map with additional functionality for dealing with PyTorch data structures. PyTorch has a lot of logic
    in the background to keep track of Parameters and gradients and where objects are located in memory. The PyTorch_Component_Map has
    a modified __setitem__ method which ensures that there are no conflicts with any of these background operations by PyTorch.
    In particular, a PyTorch_Component_Map can have a (PyTorch) Model assigned to it, and whenever __setitem__ is called, the item is
    1) Converted to a torch.nn.Parameter object if possible. This is essential for computing gradients and training the parameter.
    2) Recursively assigned if necessary. This concept is best explained with an example. Say you have a neural network with a convolutional
       layer,

       ::

            model = some_pytorch_model()
            model.conv1 = torch.nn.Conv2d(4,4,4) # This represents a 4x4 convolutional layer with 4 channels.

        'model.conv1' is itself a PyTorch Module with its own internal state, and in general, models can have models that have models,
        and so on. In other words, 'model.conv1' could itself have variables that are Modules/Models and so on. When you get the state
        dict for the original model, you will get nested dictionaries. These can still be serialized and saved to a file like normal, but
        when we call set_state, we want to make sure that we assign these nested dictionary elements to the correct submodules.

        ::

            state = model.get_state()
            internal = state['internal']
            internal['conv1'] == {'weights': ['This is some Tensor'], 'bias': ['This is some vector']}

        If we naively called model.set_state(state) to load some other state from a file, then we would end up assigning a nested
        dictionary to the value of model.conv1. What we actually want is:

        ::

            model.set_state(state)
            print(model.conv1) # This is a PyTorch Module
            print(model.conv1.weights) # This is a Tensor
            print(model.conv1.bias) # This is a Tensor

        PyTorch_Component_Map checks if the attribute being assigned to is a PyTorch_Model or (PyTorch) Module and performs this type
        of assignment.

    3) 'Registered' to the Model. This is something that PyTorch does whenever you assign a value to a PyTorch Module and is essential
        for proper functioning of PyTorch methods/functions, such as getting a state_dict, submodules, etc.

    This additional logic is important, because in general, all of the layers of a Neural Network are implemented as Modules and PyTorch
    Modules inherently has a nested structure.
    """

    def __init__(self, components, model=None):
        """
        If a model is provided, then the PyTorch_Component_Map can register values with that model, which is essential for proper usage with
        PyTorch Modules.
        """
        self.model = model
        Component_Map.__init__(self, components)

    def __setitem__(self, key, val):
        """
        This method has additional logic to __setitem__ that is described above.
        """
        # This allows the Module superclass to register the parameters.
        if self.model is not None and key != 'components' and hasattr(self.model, key) and isinstance(val, dict): # If setting the state dict for a submodule.
                submodule = getattr(self.model, key)
                if hasattr(submodule, 'set_state'): # Is a Model
                    submodule.set_state(val)
                elif isinstance(submodule, Module):
                    # Convert state dict to a dict of tensors
                    val = {k:torch.Tensor(v) for k,v in val.items()}
                    submodule.load_state_dict(val)
                elif isinstance(submodule, dict): # It was supposed to be a dict
                    Component_Map.__setitem__(self, key, val)
                    if self.model is not None:
                        val = self[key]
                        i = self.model._flags['components_initialized']
                        self.model._flags['components_initialized'] = 0
                        setattr(self.model, key, val)
                        self.model._flags['components_initialized'] = i
        else:
            Component_Map.__setitem__(self, key, val)
            if self.model is not None:
                val = self[key]
                i = self.model._flags['components_initialized']
                self.model._flags['components_initialized'] = 0
                setattr(self.model, key, val)
                self.model._flags['components_initialized'] = i

    def hook(self, key, value):
        """
        This is used to (try to) convert objects to torch.nn.Parameter objects upon assignment. If a value has tensor-like structure (ie. is
        a list or ndarray), then it will automatically be converted.
        """

        if not isinstance(value, Parameter) and not isinstance(value, Module) and hasattr(value, '__len__'): # Convert to Parameter
            try:
                value = Parameter(torch.Tensor(value))
            except:
            # If the component is not a tensor-like, Parameter, or Module, then it is some other object that we simply attach to the model
            # For example, it could be a Pipe or Junction that the model can call upon.
                pass
        return key, value

    def setitem_hook(self, key, value):
        """
        This assigns the above hook to setitem_hook so that it will be triggered upon every __setitem__ call.
        """
        return self.hook(key, value)



class EndHyperparameterOptimization(RuntimeError):
    """
    This exception can be raised to signal a factory to stop looping.
    """
    pass

class ParameterizationError(KeyError):
    """
    This exception is raised to indicate that a Model is missing required parameters that it needs to function.
    """
    pass 



import torch
from fireworks import Message, PyTorch_Model
from fireworks.extensions import IgniteJunction
from fireworks.toolbox import ShufflerPipe, BatchingPipe, TensorPipe
from fireworks.toolbox.preprocessing import train_test_split, Normalizer

import numpy as np
from random import randint

def generate_data(n=1000):

    a = randint(-10,10)
    b = randint(-10,10)
    c = randint(-10,10)
    errors = np.random.normal(0, .15, n)
    x = np.random.rand(n) * 100 - 50
    y = a + b*x + c*x**2 + errors

    return Message({'x': x, 'y': y, 'errors': errors}), {'a': a, 'b': b, 'c': c}

# Construct data, split into train/eval/test, and get get_minibatches
def get_data(n=1000):

    data, params = generate_data(n)
    train, test = train_test_split(data, test=.25)

    shuffler = ShufflerPipe(train)
    minibatcher = BatchingPipe(shuffler, batch_size=25)
    train_set = TensorPipe(minibatcher, columns=['x','y'])

    test_set = TensorPipe(test, columns=['x','y'])

    return train_set, test_set, params

class NonlinearModel(PyTorch_Model):

    required_components = ['a','b', 'c', 'd', 'e']

    def init_default_components(self):

        for letter in ['a', 'b', 'c', 'd', 'e']:
            self.components[letter] = torch.nn.Parameter(torch.Tensor(np.random.normal(0,1,1)))

        self.in_column = 'x'
        self.out_column = 'y_pred'

    def forward(self, message):

        x = message[self.in_column]
        message[self.out_column] = (self.a + self.b*x + self.c*x**2 + self.d*x**3 + self.e*x**4)

        return message



from sqlalchemy import Table, Column, Float, Integer, create_engine
from fireworks.extensions.database import create_table, TablePipe, DBPipe
from fireworks.extensions import SQLFactory
from fireworks.extensions import Experiment
import os

from model_selection import loss, train_set, test_set, params, get_trainer, Parameterizer, AccuracyMetric

if __name__=="__main__":

    description = "Model selection for nonlinear regression. We are comparing the regression accuracy of different polynomial models fit to data generated by a random polynomial."
    experiment = Experiment("model_selection_db", db_path=".", description=description)
    # SQL factory
    params_table = create_table('params', columns=[
        Column('a', Integer), Column('b', Integer), Column('c', Integer), Column('d', Integer), Column('e', Integer)
        ])
    metrics_tables = {'accuracy': create_table('accuracy', columns=[Column('average-loss', Float)])}
    # engine = create_engine('sqlite:///model_selection.sqlite')
    engine = experiment.get_engine('factory.sqlite')
    factory = SQLFactory(components={
        'trainer': get_trainer(train_set, loss, optimizer='Adam', lr=.1),
        'eval_set': test_set,
        'parameterizer': Parameterizer(),
        'metrics': {'accuracy': AccuracyMetric()},
        'engine': engine,
        'params_table': params_table,
        'metrics_tables': metrics_tables,
        })

    factory.run()

    params_table = DBPipe('params', factory.engine)
    print(params_table.all())
    accuracy_table = DBPipe('accuracy', factory.engine)
    print(accuracy_table.all())



#%%

from fireworks import PyTorch_Model, Message, HookedPassThroughPipe, Experiment
from fireworks.toolbox import ShufflerPipe, TensorPipe, BatchingPipe, FunctionPipe
from fireworks.toolbox.preprocessing import train_test_split
from fireworks.extensions import IgniteJunction
from fireworks.core import PyTorch_Model
import pandas as pd
import torch
from torchvision.datasets.mnist import FashionMNIST
from os import environ as env
from itertools import count
import matplotlib.pyplot as plt 
import visdom

env_name = 'mnist_fashion'
# vis = visdom.Visdom(env=env_name) # If you have a running Visdom server, you can uncomment this to generate plots.
description = "Here, we will train a convolutional neural network on the Fashion MNIST dataset to demonstrate the usage of Fireworks."
experiment = Experiment(env_name, description=description)
#%%

mnist_dir = env.get('MNIST_DIR', './MNIST/')
print(mnist_dir)

# First, we download our dataset and plot one of its elements as an example.
mnist = FashionMNIST(mnist_dir, download=True)
dataset = Message({'examples': mnist.data, 'labels': mnist.targets})
example = dataset['examples'][0]
plt.imshow(example)
plt.show()

#%%
# Now we construct our training and test sets as a pipeline.
train, test = train_test_split(dataset, test=.1)

# We can compose pipes to create an input pipeline that will shuffle the training set on each iteration and produce minibatches formatted for our image classifier.
shuffler = ShufflerPipe(train)
minibatcher = BatchingPipe(shuffler, batch_size=100)
to_cuda = TensorPipe(minibatcher, columns=['examples', 'labels']) # By default, all columns will be moved to Cuda if possible, but you can explicitly specify which ones as well

def tensor_to_float(batch, column='examples'):
    """ This converts the images from bytes to floats which is the data type that torch.nn.Conv2d expects. """
    batch[column] = batch[column].float()
    return batch
 
def reshape_batch(batch, column='examples'):
    """ This reshapes the batch to have an extra dimension corresponding to the input channels so we can apply the torch.nn.Conv2d operation in our model. """
    shape = batch[column].shape
    new_shape = torch.Size([shape[0], 1, shape[1], shape[2]])
    batch[column] = batch[column].reshape(new_shape)
    return batch

def normalize_batch(batch, column='examples'):
    """ Normalizes pixel intensities to fall between 0 and 1. """
    batch[column] /= 255. 
    return batch

to_float = FunctionPipe(input=to_cuda, function=tensor_to_float)
normalized = FunctionPipe(input=to_float, function=normalize_batch)
training_set = FunctionPipe(input=normalized, function=reshape_batch)

# We can also compose a pipeline in one go like we do here for the test set.
test_set = \
    FunctionPipe(
        input=FunctionPipe(
            input=FunctionPipe(
                input=TensorPipe(
                    input=BatchingPipe(
                        input=test,
                        batch_size=100
                        ), 
                    columns=['examples', 'labels']
                ), 
                function=to_float
            ),
            function=normalize_batch
        ),
        function=reshape_batch
    )

#%%

# Construct Model
class mnistModel(PyTorch_Model):
    """ Embeds each image into a 10-dimensional vector. """
    required_components = ['in_column', 'out_column', 'conv1', 'pool1', 'conv2', 'pool2']

    def init_default_components(self):

        self.components['in_column'] = 'examples'
        self.components['out_column'] = 'embeddings'
        self.components['conv1'] = torch.nn.Conv2d(1, 64, 2, padding=1)
        self.components['pool1'] = torch.nn.MaxPool2d(2)
        self.components['conv2'] = torch.nn.Conv2d(64, 32, 2)
        self.components['pool2'] = torch.nn.MaxPool2d(2)
        self.components['nonlinearity'] = torch.nn.ELU()

    def forward(self, batch):
        
        embedding = batch[self.in_column]
        embedding = self.nonlinearity(self.conv1(embedding))
        embedding = self.pool1(embedding)
        embedding = self.nonlinearity(self.conv2(embedding))
        embedding = self.pool2(embedding)
        embedding = embedding.reshape(len(batch), 1152)
        batch[self.out_column] = embedding
        return batch

class Classifier(PyTorch_Model):
    """ Uses the input embedding to perform a classification. """    
    required_components = ['in_column', 'out_column', 'linear_layer']

    def init_default_components(self):
        self.components['in_column'] = 'embeddings'
        self.components['out_column'] = 'predictions'
        self.components['linear1'] = torch.nn.Linear(1152, 256)
        self.components['linear2'] = torch.nn.Linear(256, 10)
        self.components['nonlinearity'] = torch.nn.ELU()
        self.components['softmax'] = torch.nn.Softmax(dim=1)

    def forward(self, batch):
        
        predictions = batch[self.in_column]
        predictions = self.nonlinearity(self.linear1(predictions))
        predictions = self.softmax(self.linear2(predictions))        
        batch[self.out_column] = predictions
        return batch 

# All function calls to the classifier will call the embedder first
# ie. classifier(x) is equivalent to classifier.forward(embedder.forward(x))
embedder = mnistModel()
classifier = Classifier(input=embedder) 

if torch.cuda.is_available():
    embedder.cuda()
    classifier.cuda()
#%%

# Set up loss function and training loop
ce_loss = torch.nn.CrossEntropyLoss()
loss = lambda batch: ce_loss(batch['predictions'], batch['labels'])

# By default, this Junction applies a standard training closure of evaluating the model,
# computing gradients of the loss, and backpropagating using the chosen optimizer.
trainer = IgniteJunction(
    components={
        'model': classifier, 
        'dataset': training_set
        }, 
    loss=loss, optimizer='Adam', 
    lr=.0001, weight_decay=.001, 
    visdom=False, # If you have a running Visdom server, you can set this to true to plot training loss over time.
    environment=env_name
    )

trainer.run(max_epochs=10) # This will take almost 20 minutes on CPU and around 1 minute on GPU

#%%
# Now that we've trained our model, we can compute some metrics on the test set.
# Here, we construct a Pipe that will compute metrics such as sensitivity, specificity, f1, etc.
# on the test set.

classes = {i: class_name for i, class_name in zip(count(), mnist.classes)}
class Metrics(HookedPassThroughPipe):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.true_positives = {class_name: 0 for class_name in classes.values()}
        self.true_negatives = {class_name: 0 for class_name in classes.values()}
        self.false_positives = {class_name: 0 for class_name in classes.values()}
        self.false_negatives = {class_name: 0 for class_name in classes.values()}
        self.total_count = 0
        self.label_counts = {class_name: 0 for class_name in classes.values()}
        self.prediction_counts = {class_name: 0 for class_name in classes.values()}

    def _call_hook(self, batch):
        """ 
        This will get called every time the model is called. As a result, this pipe will continuously update
        itself as we iterate through the test set.
        """
        labels = batch['labels']
        predictions = torch.max(batch['predictions'],1)[1]
        correct_indices = (predictions == labels).nonzero().flatten().tolist()
        incorrect_indices = (predictions != labels).nonzero().flatten().tolist()
        for index, name in classes.items():
            self.label_counts[name] += int(sum(labels == index)) # How often the class showed up
            self.prediction_counts[name] += int(sum(predictions == index)) # How often the class was predicted
            self.true_positives[name] += int(sum(predictions[correct_indices] == index)) # How often the correct prediction was for thsi class
            self.true_negatives[name] += int(sum(predictions[correct_indices] != index)) # How often the correct prediction was not for the class; ie. how often the prediction was a true negative for this class
            self.false_positives[name] += int(sum(predictions[incorrect_indices] == index)) # How often a wrong prediction was for this class
            self.false_negatives[name] += int(sum(predictions[incorrect_indices] != index)) # How often a wrong prediction was for another class; ie. how often the prediction was a false negative for this class
        self.total_count += len(batch)
        return batch

    def compile_metrics(self):
        """ 
        After we have gone through the entire test set, we can call this method to compute the actual metrics.
        """
        class_names = classes.values()
        negative_counts = {name: sum(self.label_counts[other] for other in class_names if other != name) for name in class_names}
        self.sensitivity = {name: self.true_positives[name] / self.label_counts[name] for name in class_names}
        self.specificity = {name: self.true_negatives[name] / negative_counts[name] for name in class_names}
        negative_prediction_counts = {name: sum(self.prediction_counts[other] for other in class_names if other != name) for name in class_names}
        self.ppv = {name: self.true_positives[name] / self.prediction_counts[name] for name in class_names}
        self.npv = {name: self.true_negatives[name] / negative_prediction_counts[name] for name in class_names}
        self.f1 = {name: 2 / (1/self.ppv[name] + 1/self.sensitivity[name]) for name in class_names}
        self.accuracy = {name: (self.true_positives[name] + self.true_negatives[name]) / self.total_count for name in class_names}

    def get_metrics(self):
        """
        Lastly, we will use this method to return the computed metrics as a Pandas DataFrame.
        """
        columns = ['sensitivity', 'specificity', 'ppv', 'npv', 'f1', 'accuracy']
        df = pd.DataFrame(columns=columns, index=classes.values())
        for attribute in columns:
            value = getattr(self, attribute)
            df[attribute] = [value[key] for key in df.index]

        return df

# This class is implemented a a HookedPassThroughPipe, meaning that it's _call_hook method will be applied every time
# The class is called like a function, and this call will pass through to its input.
metrics_computer = Metrics(input=classifier)

for batch in test_set:
    # We can simply call this object repeatedly on batches in the test set
    # This operation is equivalent to metrics_computer._call_hook(classifier(batch))
    metrics_computer(batch)

metrics_computer.compile()
df = metrics_computer.get_metrics()
print(df)

# You can also convert this DataFrame to a Message.
m = Message(df)
print(m)

# Lastly, we can save our results from this experiment
# At it's simplest, the experiment object gives you a way of organizing files.
# You can uses its open() method to get a file handle or path string to a file 
# inside its experiment directory.
# Each time you create an experiment, a new experiment directory will be created automatically.
df.to_csv(experiment.open("metrics.csv", string_only=True))
# Since our models are still subclasses of torch.nn.module, we can save them using the standard torch.save feature
# but if we want, we can also save their parameters in other formats such as JSON
state = embedder.get_state()
Message.from_objects(state).to('json', path=experiment.open("embedder.json",string_only=True))
state = classifier.get_state()
Message.from_objects(state).to('json', path=experiment.open("classifier.json",string_only=True))



from abc import ABC, abstractmethod
from Bio import SeqIO
import pandas as pd
from .message import Message
from fireworks.utils import index_to_list
from fireworks.utils.exceptions import ParameterizationError
from .component_map import Component_Map, PyTorch_Component_Map
from .cache import LRUCache, LFUCache, UnlimitedCache
from abc import ABC, abstractmethod
from itertools import count
import types
import random
from bidict import bidict
import torch
import math

class Junction:
    """
    A junction can take pipes as inputs, and its outputs can be piped to other pipes.
    All communication is done via Message objects.

    Unlike Pipes, junctions do not automatically have recursive method calling. This is because they have multiple input sources,
    which would result in ambiguity. Instead, junctions are meant to act as bridges between multiple pipes in order to enable
    complex workflows which require more than a linear pipeline.

    Like Models, Junctions can have internal and external components in their state. 
    """

    def __init__(self, *args, components=None, **kwargs):

        self.components = Component_Map(components)
        self.check_components()

    def check_components(self, components = None):
        """
        Checks to see if the provided components dict provides all necessary params for this model to run.
        """
        if components is None:
            components = self.components
        missing_components = []
        error = False
        for key in self.required_components:
            if key not in components:
                missing_components.append(key)
                error = True
        if error:
            raise ParameterizationError("Missing required components {0}".format(missing_components))

    def type_check(self, key, components):

        if type(self.required_components) is dict and key in self.required_components:
            if not isinstance(components[key], self.required_components[key]):
                # raise ValueError("Component {0} with value {1} is not the correct type. Must be an instance of {2}".format(key, components[key], self.required_components[key))
                raise ValueError
        else:
            return True

    def __setattr__(self, name, value):

        if name.startswith('__') or name == 'components':
            object.__setattr__(self, name, value)
        else:
            self.type_check(name, value)
            self.components[name] = value

    def __getattr__(self, name):

        if name in self.components:
            return self.components[name]
        else:
            raise AttributeError

    @property
    def required_components(self):
        """
        This should be overridden by a subclass in order to specify components that should be provided during initialization. Otherwise,
        this will default to just return the components already present within the Model.
        """
        return self.components.keys()

    def set_state(self, state, *args, **kwargs): self.components.set_state(state)

    def get_state(self): return self.components.get_state()

    def save(self, *args, **kwargs):

            save_dict = self._save_hook(*args, **kwargs)
            if save_dict == {}:
                pass
            else:
                save_df = Message.from_objects(save_dict).to_dataframe().df

            for name, component in self.components.items():
                component.save(*args, **kwargs)

    def _save_hook(self):

        return {}

class PyTorch_Junction(Junction):
    """
    A PyTorch Junction can have automatically convert components to their PyTorch equivalents (eg. convert a numpy array or list to a
    torch.Tensor), and this can be useful when using the Junction for PyTorch related tasks.
    """
    def __init__(self, *args, components=None, **kwargs):
        self.components = PyTorch_Component_Map(components)
        self.check_components()



from sqlalchemy import Table, Column, Integer, String, create_engine
from fireworks import Message
from fireworks.extensions import database as db
from fireworks import pipe as pl
import os
import numpy as np
import itertools
import copy

class dummy_pipe(pl.Pipe):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.length = 20

    def __getitem__(self, index):

        if type(index) is list:
            index = [i for i in index]
        if type(index) is slice:
            step = index.step or 1
            index = [i for i in range(index.start, index.stop, step)]

        if index == []:
            return None
        elif max(index) < self.length and min(index) >= 0:
            return Message({'name': 'johnny', 'values': np.array(index)})
        else:
            raise IndexError("Out of bounds for dummy pipe with length {0}.".format(self.length))

    def __len__(self):
        return self.length

def dummy_table(table_name):

    columns = [
        Column('name', String),
        Column('values', Integer),
        ]
    tab = db.create_table(table_name, columns)

    return tab

def test_make_row():

    tab = dummy_table('bubsy')
    tom = tab(name='ok', values=33)
    assert tom.name == 'ok'
    assert tom.values == 33

    engine = create_engine('sqlite:///:memory:', echo=True)
    pipe = db.TablePipe(tab, engine)

    message = Message({'name': ['a','b'], 'values': [1,2]})
    row = pipe.make_row(message[0])
    assert row.name == 'a'
    assert row.values == 1

def test_create_table():

    tab = dummy_table('munki')
    assert tab.__table__.name == 'munki'
    for colname in ['name', 'values', 'id']:
        assert hasattr(tab, colname)

def test_parse_columns():

    tab = dummy_table('joseph')
    colnames = db.parse_columns(tab)
    for c in colnames:
        assert type(c) is str
    assert set(['values', 'name']) == set(colnames)
    colnames = db.parse_columns(tab, ignore_id=False)
    for c in colnames:
        assert type(c) is str

    assert set(['values', 'id', 'name']) == set(colnames)

def test_TablePipe_explicit():
    """ Colnames are explicitly labeled here. """

    dummy = dummy_pipe()
    tab = dummy_table('jojo')
    engine = create_engine('sqlite:///:memory:', echo=True)

    tab.metadata.create_all(engine)
    ts = db.TablePipe(tab, engine, ['values', 'name'], input=dummy)
    batch = ts[2:10]
    ts.insert(batch)
    ts.commit()
    # Check if it worked
    for row, i in zip(ts.query(), itertools.count()):
        assert type(row) is Message
        assert row['name'][0] == 'johnny'
        # assert int.from_bytes(row.values, byteorder='little') == i+2 # Have to convert integers back from little endian
        assert row['values'][0] == i+2
    assert i > 1

def test_TablePipe_implicit():
    """ Colnames are implicitly identified here. """

    dummy = dummy_pipe()
    tab = dummy_table('josuke')
    engine = create_engine('sqlite:///:memory:', echo=True)

    ts = db.TablePipe(tab, engine, input=dummy)
    batch = ts[2:10]
    ts.insert(batch)
    ts.commit()
    # Check if it worked
    for row, i in zip(ts.query(), itertools.count()):
        assert type(row) is Message
        assert row['name'][0] == 'johnny'
        # assert int.from_bytes(row.values, byteorder='little') == i+2 # Have to convert integers back from little endian
        assert row['values'][0] == i+2

    # Test deletes
    cop = ts.query().filter('values','between',4,8).all()
    assert len(cop) == 5
    assert (ts.query().all()['values'] == [2,3,4,5,6,7,8,9]).all()
    ts.delete('values', batch['values'][1:4])
    ts.commit()
    assert (ts.query().all()['values'] == [2,6,7,8,9]).all()
    assert 'id' in ts.query().all()

    # Test updates
    batch = ts.query().all()
    # assert False
    new_batch = copy.deepcopy(batch)
    new_batch['values'] = [10,11,12,13,14]
    ts.update('id', new_batch)
    ts.commit()
    newer_batch = ts.query().all()
    assert newer_batch == new_batch
    assert newer_batch != batch

def test_DBPipe():
    dummy = dummy_pipe()
    tab = dummy_table('jotaro')
    engine = create_engine('sqlite:///jotaro.sqlite', echo=True)
    if os.path.exists('jotaro.sqlite'):
        os.remove('jotaro.sqlite')
    tab.metadata.create_all(engine)
    ts = db.TablePipe(tab, engine, ['name', 'values'], input=dummy)
    batch = ts[2:10]
    ts.insert(batch)
    ts.commit()
    deedee = db.DBPipe(tab, engine)
    for row, i in zip(deedee, itertools.count()):
        assert row == Message({'id':[i+1],'name': ['johnny'], 'values':[i+2]})
    deedee.reset_session()
    for row, i in zip(deedee, itertools.count()):
        assert row == Message({'id':[i+1],'name': ['johnny'], 'values':[i+2]})

    # Test using reflections
    ts = db.DBPipe('jotaro', engine)
    for row, i in zip(deedee, itertools.count()):
        assert row == Message({'id':[i+1],'name': ['johnny'], 'values':[i+2]})
    deedee.reset_session()
    for row, i in zip(deedee, itertools.count()):
        assert row == Message({'id':[i+1],'name': ['johnny'], 'values':[i+2]})

    # Reset the session and try again.
    deedee.reset_session()
    for row, i in zip(deedee, itertools.count()):
        assert row == Message({'id':[i+1],'name': ['johnny'], 'values':[i+2]})
    deedee.reset_session()
    for row, i in zip(deedee, itertools.count()):
        assert row == Message({'id':[i+1],'name': ['johnny'], 'values':[i+2]})

    # Test using reflections
    ts = db.DBPipe('jotaro', engine)
    for row, i in zip(deedee, itertools.count()):
        assert row == Message({'id':[i+1],'name': ['johnny'], 'values':[i+2]})
    deedee.reset_session()
    for row, i in zip(deedee, itertools.count()):
        assert row == Message({'id':[i+1],'name': ['johnny'], 'values':[i+2]})

def test_DBPipe_query():

    dummy = dummy_pipe()
    tab = dummy_table('giorno')
    engine = create_engine('sqlite:///:memory:', echo=True)
    tab.metadata.create_all(engine)
    ts = db.TablePipe(tab, engine, ['name', 'values'], input=dummy)
    batch = ts[0:20]
    ts.insert(batch)
    ts.commit()
    result = ts.query().all()
    result = ts.query('values').all()
    assert set(result.columns) == set(['values'])
    result = ts.query('values', 'name').all()
    assert set(result.columns) == set(['name', 'values'])
    result = ts.query('values').filter('values', 'between', 1,5).all()
    assert (result['values'] == [1,2,3,4,5]).all()
    assert set(result.columns) == set(['values'])
    result = ts.query('values', 'name').filter('values', 'between', 1,5).all()
    assert (result['values'] == [1,2,3,4,5]).all()
    assert set(result.columns) == set(['name', 'values'])

    # Test if resetting the session after making a query affects anything.
    query = ts.query('values')
    query.reset_session()
    result = query.all()
    assert set(result.columns) == set(['values'])
    query = ts.query('values', 'name')
    query.reset_session()
    result = query.all()
    assert set(result.columns) == set(['name', 'values'])
    query = ts.query('values').filter('values', 'between', 1,5)
    query.reset_session()
    result = query.all()
    assert (result['values'] == [1,2,3,4,5]).all()
    assert set(result.columns) == set(['values'])
    query = ts.query('values', 'name')
    query.reset_session()
    result = query.filter('values', 'between', 1,5).all()
    assert (result['values'] == [1,2,3,4,5]).all()
    assert set(result.columns) == set(['name', 'values'])

def test_reflect_table():

    tab = dummy_table('jolyne')
    engine = create_engine('sqlite:///:jolyne.sqlite', echo=True)
    tab.metadata.create_all(engine)
    jolyne = db.reflect_table('jolyne', engine)
    assert type(jolyne) is Table
    assert jolyne.columns.keys() == tab.__table__.columns.keys()



from . import cache, message, junction, pipe, model
from .message import Message, TensorMessage, cat
from .pipe import Pipe, HookedPassThroughPipe, recursive
from .model import Model, PyTorch_Model, model_from_module
from .junction import Junction



import numpy as np

"""
This file contains helper functions that can be useful for NLP tasks.
"""

def character_tokenizer(sequence):
    """ Splits sequence into a list of characters. """

    if sequence:
        return [*sequence] # This unpacks an iterable into a list of elements

def space_tokenizer(sequence):
    """ Splits sequence based on spaces. """
    if sequence:
        return sequence.split(' ')

def pad_sequence(sequence, max_length, embeddings_dict, pad_token = 'EOS'):
    """ Adds EOS tokens until sequence length is max_length. """

    n = len(sequence)
    padded = sequence.copy()

    if n < max_length:
        if type(padded) is np.ndarray:
            padded = np.append(padded, [embeddings_dict[pad_token] for _ in range(max_length-n)], axis=0)
        else:
            padded.extend([embeddings_dict[pad_token] for _ in range(max_length-n)])

    return np.array(padded)

def pad(batch, embeddings_dict, pad_token = 'EOS'): # TODO: Don't rely on embeddings dict for EOS tokens
    """ Pads all embeddings in a batch to be the same length. """

    lengths = [len(x) for x in batch]
    max_length = max(lengths)
    padded_embeddings = np.stack([pad_sequence(seq, max_length, embeddings_dict, pad_token) for seq in batch])
    batch = padded_embeddings

    return batch

def apply_embeddings(sequence, embeddings_dict, tokenizer):
    """ Decomposes sequence into tokens using tokenizer and then converts tokens to embeddings using embeddings_dict. """

    if sequence:
        embeddings = [embeddings_dict['SOS']]
        embeddings.extend([embeddings_dict[token] if token in embeddings_dict else embeddings_dict['UNK']
                        for token in tokenizer(sequence)])
        embeddings.append(embeddings_dict['EOS'])
        # if vectorizor:
        #     embeddings = [vectorizor(embedding) for embedding in embeddings]
        return embeddings

    else: # In case input sequence is empty
        return [embeddings_dict['SOS'], embeddings_dict['EOS']]

def create_pretrained_embeddings(embeddings_file): # TODO: Write test
    """ Loads embeddings vectors from file into a dict. """
    embeddings_dict = {}
    with open(embeddings_file) as vectors:
        for line in vectors:
            splits = line.split(' ')
            embeddings_dict[splits[0]] = np.array([float(x) for x in splits[1:]])

    return embeddings_dict

def load_embeddings(name = 'glove840b'): # TODO: write test
    """ Loads serialized embeddings from pickle. """
    with open(os.path.join(silph.RAW_DIR, name), 'rb') as pretrained:
        embeddings_dict = pickle.load(pretrained)

    return embeddings_dict

def make_vocabulary(text, tokenizer = None, cutoff_rule = None): #TODO: write test
    """ Converts an iterable of phrases into the set of unique tokens that are in the vocabulary. """
    counts = defaultdict(lambda:0)
    if tokenizer is None:
        tokenizer = space_tokenizer

    #vocabulary = set()
    for phrase in text:
        if phrase is not None:
            tokens = tokenizer(phrase)
            #vocabulary.add(tokens)
            for token in tokens:
                counts[token] += 1
        else:
            pass

    if cutoff_rule:
        counts = cutoff_rule(counts)

    return dict(counts) # The keys of this dict are the unique tokens.

def make_indices(vocabulary): # TODO: write test
    """
    Constructs a dictionary of token names to indices from a vocabulary.
    Each index value corresponds to a one-hot vector.
    """
    embeddings_dict = bidict({token: i for token, i in zip(vocabulary, count(start=3))})
    embeddings_dict['SOS'] = special_tokens['SOS']
    embeddings_dict['EOS'] = special_tokens['EOS']
    embeddings_dict['UNK'] = special_tokens['UNK']
    return embeddings_dict

def too_big(dataset, start, end, dim = 300, cutoff = 620000):
    """
    Calculates if a batch consisting of dataset[start:end] is too big based on cutoff. This can be used for constructing dynamic batches.
    """

    sizes = [len(x) for x in dataset[start:end]['embeddings']]
    max_size = max(sizes)
    dim = dim
    batch_size = end-start
    size = dim*max_size*batch_size
    return size > cutoff



from fireworks import Message
from fireworks import experiment as exp
from fireworks import database as db
from sqlalchemy import Column, Integer, String
from shutil import rmtree
import datetime
import pandas as pd
import os

columns = [
    Column('superpower', String),
    Column('name', String),
    Column('age', Integer),
]
table = db.create_table('superpowers', columns=columns)

def dummy_table(engine):

    return db.TablePipe(table, engine, columns=['superpower', 'name', 'age'])

def test_Experiment():

    # Test directory creation
    dirs = os.listdir()
    all_xmen = [x for x in dirs if x.startswith('xmen')]
    for man in all_xmen:
        rmtree(man)
    dirs = os.listdir()
    all_xmen = [x for x in dirs if x.startswith('xmen')]
    assert len(all_xmen) == 0
    xmen = exp.Experiment('xmen', os.getcwd(), description='ok')
    xmen2 = exp.Experiment('xmen', os.getcwd(), description='yahoo')
    dirs = os.listdir()
    all_xmen = [x for x in dirs if x.startswith('xmen')]
    assert set(all_xmen) == set(['xmen_0', 'xmen_1'])

    # Test metadata table
    query = xmen.metadata.query()
    i = 0
    for row in query:
        i += 1
    assert i == 1
    assert row['name'][0] == 'xmen'
    assert row['iteration'][0] == 0
    assert row['description'][0] == 'ok'
    assert type(row['timestamp'][0]) is pd.Timestamp

    # Test engine/session creation
    assert xmen.engines == {}
    nginx = xmen.get_engine('wolverine')
    assert 'wolverine' in xmen.engines

    wolverine = xmen.get_session('wolverine')
    xavier = xmen.get_session('xavier')
    assert 'xavier' in xmen.engines

    # Test that the engine works
    saver = dummy_table(nginx)
    saver.insert(Message({
        'superpower': ['flying', 'walking', 'eating'],
        'name': ['flyman', 'walkergirl', 'bumbo'],
        'age': [2,3,4],
        }))
    saver.commit()
    rows = saver.query()

    for row in rows:
        assert type(row) is Message
    assert row == Message({'superpower': ['eating'], 'name':['bumbo'], 'age':[4], 'id':[3]})

    # Test file opening and saving
    path = xmen.open('ironman', string_only=True)
    assert path == os.path.join(xmen.db_path, xmen.save_path, 'ironman')
    with xmen.open('ironman', 'w') as ironman:
        ironman.write("Hey guys it's me tony the tiger.")
    # try:
    #     path = xmen.open('ironman', string_only=True)
    #     assert False
    # except IOError: # It should detect that this file already exists
    #     assert True

    with xmen.open('ironman') as ironman:
        assert ironman.read() == "Hey guys it's me tony the tiger."

    # Clean up
    all_xmen = [x for x in dirs if x.startswith('xmen')]
    for man in all_xmen:
        rmtree(man, ignore_errors=True)

def test_paths():
    """
    Tests that relative and absolute paths work with Experiment constructor
    """

    xmen = exp.Experiment('xmen', 'data')
    xbois = exp.Experiment('xbois',os.path.join(os.getcwd(),'data'))
    for folder in os.listdir('data'):
        if folder != 'README.md':
            rmtree(os.path.join('data',folder))

def test_load_experiment():
    dirs = os.listdir()
    all_avengers = [x for x in dirs if x.startswith('avenger')]
    for man in all_avengers:
        rmtree(man)
    dirs = os.listdir()
    all_avengers = [x for x in dirs if x.startswith('avenger')]
    assert len(all_avengers) == 0
    avenger = exp.Experiment('avenger', os.getcwd(), description='ok')
    ironman = avenger.get_engine('ironman')
    saver = dummy_table(ironman)
    saver.insert(Message({
        'superpower': ['flying', 'walking', 'eating'],
        'name': ['flyman', 'walkergirl', 'bumbo'],
        'age': [2,3,4],
        }))
    saver.commit()
    marvel = exp.load_experiment(os.path.join(avenger.db_path,avenger.save_path))
    saver = dummy_table(ironman)
    rows = saver.query()
    for row in rows:
        assert type(row) is Message
    assert row == Message({'superpower': ['eating'], 'name':['bumbo'], 'age':[4], 'id':[3]})
    assert marvel.name == avenger.name
    assert marvel.iteration == avenger.iteration
    assert marvel.description == avenger.description
    assert marvel.timestamp == avenger.timestamp



from fireworks.core.scaffold import Scaffold
from fireworks.utils.test_helpers import *
import os
from shutil import rmtree

test_dir = 'scaffold_test'
epsilon = .001

def test_Scaffold():
    D = RandomJunction(components={})
    C = reset_dummy()
    A = DummyModel({'m': [3.]}, out_column='y1', input=C)
    A.conv1 = torch.nn.Conv2d(4,4,4)
    B = DummyModel({'m': [1.], 'b': [2.]}, input=A, in_column='y1', out_column='y')
    cafe = Scaffold({'a':A, 'b': B, 'c': C, 'd': D})
    state = cafe.serialize()
    for key in ['a','b','c','d']:
        assert 'internal' in state[key]
        assert 'external' in state[key]

    try:
        rmtree(test_dir)
    except:
        pass

    os.mkdir(test_dir)
    cafe.save(path=test_dir, method='json')

    D2 = RandomJunction(components={})
    C2 = reset_dummy()
    A2 = DummyModel({'m': [4.]}, out_column='y1', input=C)
    A2.conv1 = torch.nn.Conv2d(4,4,4)
    B2 = DummyModel({'m': [5.], 'b': [6.]}, input=A, in_column='y3', out_column='y4')
    cafe2 = Scaffold({'a':A2,'b':B2,'c':C2,'d':D2})
    assert (A2.m == 4).all()
    assert (B2.m == 5).all()
    assert (B2.b == 6).all()
    assert not (A2.conv1.state_dict()['bias'] == A.conv1.state_dict()['bias']).all()
    assert not (A2.conv1.state_dict()['weight'] == A.conv1.state_dict()['weight']).all()
    assert B2.in_column == 'y3'
    assert B2.out_column == 'y4'
    cafe2.load(test_dir)
    assert (A2.m == 3).all()
    assert (B2.m == 1).all()
    assert (B2.b == 2).all()
    assert (A2.conv1.state_dict()['bias'] - A.conv1.state_dict()['bias'] < epsilon).all()
    assert (A2.conv1.state_dict()['weight'] - A.conv1.state_dict()['weight'] < epsilon).all()
    assert B2.in_column == 'y1'
    assert B2.out_column == 'y'
    rmtree(test_dir)



from fireworks.core.component_map import Component_Map, PyTorch_Component_Map
import torch

class DummyClass:

    def __init__(self):
        self.x = 1

def test_setitem_delitem():

    dummy = DummyClass()
    mapper = Component_Map({'a': 2, 'b': 3, 'c': (dummy,'x')})
    assert mapper['a'] == 2
    assert mapper['b'] == 3
    assert mapper['c'] == 1
    assert 'a' in mapper._internal_components and 'a' not in mapper._external_modules and 'a' not in mapper._external_attribute_names
    assert 'b' in mapper._internal_components and 'b' not in mapper._external_modules and 'b' not in mapper._external_attribute_names
    assert 'c' not in mapper._internal_components and 'c' in mapper._external_modules and 'c' in mapper._external_attribute_names
    dummy.x = 2
    assert mapper['c'] == 2
    mapper['d'] = (4, 3)
    assert mapper['d'] == (4,3)
    assert 'd' in mapper._internal_components and 'd' not in mapper._external_modules and 'd' not in mapper._external_attribute_names
    new_dummy = DummyClass()
    mapper['e'] = (new_dummy, 'x')
    assert mapper['e'] == 1
    assert 'e' not in mapper._internal_components and 'e' in mapper._external_modules and 'e' in mapper._external_attribute_names
    del mapper['c']
    assert 'c' not in mapper and 'c' not in mapper._external_modules and 'c' not in mapper._external_attribute_names
    del mapper['a']
    assert 'a' not in mapper and 'a' not in mapper._internal_components

    # Test overwrites from external to internal.
    mapper['c'] = 3
    assert mapper['c'] == 3
    assert 'c' in mapper._internal_components and 'c' not in mapper._external_modules and 'c' not in mapper._external_attribute_names
    mapper['a'] = (dummy, 'x')
    assert mapper['a'] == 2
    assert 'a' not in mapper._internal_components and 'a' in mapper._external_modules and 'a' in mapper._external_attribute_names


def test_pytorch_component_map():

    dummy = DummyClass()
    mapper = PyTorch_Component_Map({'a': [1.,2.,3.], 'b': 3, 'c': 'ok', 'd': (dummy, 'x')})
    assert isinstance(mapper['a'], torch.nn.Parameter)
    assert (mapper['a'] == torch.Tensor([1,2,3])).all()
    assert mapper['b'] == 3
    assert mapper['c'] == 'ok'
    assert mapper['d'] == 1
    dummy.x = 2
    assert mapper['d'] == 2
    del mapper['a']
    assert 'a' not in mapper

def test_setstate_getstate():

    dummy = DummyClass()
    mapper = Component_Map({'a': 2, 'b': 3, 'c': (dummy,'x')})
    state = mapper.get_state()
    assert state['internal']['a'] == 2
    assert state['internal']['b'] == 3
    assert state['external']['c'] == (dummy, 'x')
    new_dummy = DummyClass()
    new_state = {'internal': {'a': 4, 'c': 5}, 'external': {'b': (new_dummy, 'x')}}
    mapper.set_state(new_state)
    assert mapper['a'] == 4
    assert mapper['c'] == 5
    assert mapper['b'] == 1
    new_dummy.x = 2
    assert mapper['b'] == 2



from abc import ABC, abstractmethod, abstractproperty
import torch
from torch.nn import Module, Parameter
from abc import abstractmethod
from fireworks.utils.exceptions import ParameterizationError
from .message import Message
from .junction import Junction, PyTorch_Junction
from .component_map import Component_Map, PyTorch_Component_Map
from .pipe import HookedPassThroughPipe, recursive
import os
import pandas as pd

class Model(HookedPassThroughPipe, Junction, ABC):
    """
    Represents a statistical model which has a set of components, and a
    means for converting inputs into outputs. The model functions like a Pipe
    with respect to the input/output stream, and it functions like a Junction
    with respect to the parameterization. components can be provided via multiple
    different sources in this way, providing flexibility in model configuration.
    Models can also provide components for other Models, enabling one to create
    complex graphs of Models that can be trained simultaneously or individually.
    """

    def __init__(self, components = {}, *args, input = None, **kwargs):
        """
        Args:
            components: A dict of components that the model can call on.
        """
        self._flags = {'recursive_get': 1, 'components_initialized': 0} # Used for controlling recursion. Don't mess with this.
        self.components = Component_Map(components, owner=self)
        HookedPassThroughPipe.__init__(self, input = input)

        self.init_default_components()
        self.enable_updates()
        self.enable_inference()
        self._flags['components_initialized'] = 1

    def init_default_components(self):
        """
        This method can optionally be implemented in order for the model to provide a default initialization for some or all of its
        required components.
        """
        pass

    @abstractmethod
    def forward(self, message):
        """
        Represents a forward pass application of the model to an input. Must be implemented by a subclass.
        This should return a Message.
        """
        pass

    def save(self, *args, method='json', **kwargs):
        """
        Aggregates this model's components into a single Message and saves them using the chosen method.
        You can use any method that Messages support writing to via the to_ method, and you can
        provide additional key-word arguments as needed to support this.
        If the save method involves a path, then the path will be modified for each component and the state_dict.
        For the state_dict, the path will be torch_{name}-{path}, and for each component it will be {key}_{path}
        where name is either the name of the class or self.__name__ if it is defined. key is the value of the key in
        the components dictionary.
        """
        # Get name of Model
        if hasattr(self, '__name__'):
            name = self.__name__
        else:
            name = type(self).__name__

        # Parse save path if provided
        if 'path' in kwargs:
            kwargs = kwargs.copy()
            path = kwargs['path']
            paths = path.split('/')
            paths[-1]="{0}-{1}".format(name, paths[-1])

        state = self.get_state()
        external = state['external']
        internal = state['internal']

        # Save internal state_dict
        if 'path' in kwargs:
            kwargs['path'] = os.path.join(*paths[:-1], 'torch_'+paths[-1])

        internal_state_as_message = Message.from_objects(internal)
        serialized = internal_state_as_message.to(method=method, **kwargs)

        external_components = set()
        external_dict = {}

        for key, value in external.items():
            if value[0] not in external_components:
                external_components.add(value[0])
                external_dict[key] = value[0]

        # Save other components. This should recursively trigger the same action on the components if necesssary.
        for key, component in (external_dict.items()):
            if hasattr(component, 'save'):
                if 'path' in kwargs:
                    kwargs['path'] = os.path.join(*paths[:-1], key+'_'+paths[-1])
                component.save(method=method, **kwargs)

        # Save inputs. This is done by default and can be disabled by providing 'recursive=False' as an argument.
        if ('recursive' in kwargs and kwargs['recursive']) or 'recursive' not in kwargs:

            if 'path' in kwargs:
                new_path = paths[-1].split('.')
                new_path[0] = '--' + new_path[0]
                new_path = '.'.join(new_path)
                kwargs['path'] = os.path.join(*paths[:-1], new_path)
            try:
                self.recursive_call('save',method=method, **kwargs)
            except AttributeError: #TODO: Make a custom RecursionEnd error to indicate that the recursion has ended in order to provide tighter error checking.
                pass

        return serialized

    def load_state(self, *args, method='json', **kwargs):
        """
        Loads the data in the given save file into the state dict.
        """
        if 'path' in kwargs:
            kwargs = kwargs.copy()
            kwargs['filepath'] = kwargs['path']
            del kwargs['path']

        methods = {
            'json': pd.read_json,
            'pickle': pd.read_pickle,
            'csv': pd.read_csv,
        }

        df = methods[method](*args, **kwargs) # Load parameters in
        state_dict = Message(df).to_dict()
        state_dict = {key: value[0] for key, value in state_dict.items()}
        self.set_state({'external':{}, 'internal': state_dict}, reset=False)

    def get_state(self):
        return self.components.get_state()

    def set_state(self, state, reset=True):

        if reset:
            self.components = Component_Map({}, owner=self)
            self.init_default_components()
        self.components.set_state(state)

    def update(self, message, **kwargs): pass

    def compile(self): pass

    # TODO: Implement description methods (__repr__, __str__)

    def _getitem_hook(self, message):

        self._update_hook(message, method='get')
        return self._forward_hook(message)

    def _next_hook(self, message):

        self._update_hook(message, method='next')
        return self._forward_hook(message)

    def _call_hook(self, message, *args, **kwargs):

        self._update_hook(message, method='call')
        return self._forward_hook(message, *args, **kwargs)

    def __call__(self, *args, **kwargs):

        return HookedPassThroughPipe.__call__(self, *args, **kwargs)

    def __getattr__(self, name):

        if self._flags['components_initialized'] and name in self.components:
            return self.components[name]

        if not name.startswith('_') and self._flags['recursive_get']:
            try:
                return HookedPassThroughPipe.__getattr__(self, name)
            except:
                pass

        raise AttributeError("'{}' object has no attribute '{}'".format(
            type(self).__name__, name))

    # def __setattr__(self, name, value):
    #     """
    #     Attribute modifications ignore the recursive aspect of Pipes.
    #     """
    #
    #     if name in ['_flags', 'input']:
    #         object.__setattr__(self, name, value)
    #     else:
    #         self._flags['recursive_get'] = 0
    #         Module.__setattr__(self, name, value)
    #         self._flags['recursive_get'] = 1
    #         if self._flags['components_initialized']:
    #             # self.update_components()
    #             self.components[name] = value

    def enable_inference(self):

        self._forward_hook = self.forward
        self._inference_enabled = True

    @recursive()
    def enable_inference_all(self):
        self.enable_inference()

    def disable_inference(self): #TODO: test this

        self._forward_hook = identity
        self._inference_enabled = False

    @recursive()
    def disable_inference_all(self):
        self.disable_inference()

    def enable_updates(self):

        self._updates_enabled = True
        self._update_hook = self.update

    @recursive()
    def enable_updates_all(self):
        self.enable_updates()

    def disable_updates(self):

        self._updates_enabled = True
        self._update_hook = identity

    @recursive()
    def disable_updates_all(self):
        self.disable_updates()

def identity(message, *args, **kwargs):

    return message

def freeze_module(module, parameters = None, submodules = None):
    """
    Recursively freezes the parameters in a PyTorch module.
    """
    _change_temperature(False, module, parameters, submodules)

def unfreeze_module(module, parameters = None, submodules = None):
    """
    Recursively unfreezes the parameters in a PyTorch module.
    """
    _change_temperature(True, module, parameters, submodules)

def _change_temperature(boo, module, parameters = None, submodules = None):
    """
    Changes the temperature of a PyTorch module.
    """
    parameters = parameters or module.parameters()
    submodules = submodules or module.modules()

    for parameter in parameters:
        parameter.requires_grad = boo
    for submodule in submodules:
        if submodule is not module:
            change_temperature(boo, submodule)

def model_from_module(module_class):
    """
    Given the class definition for a pytorch module, returns a model that
    encapsulates that module.
    """
    class ModelFromModule(module_class, PyTorch_Model):

        def __init__(self, components={}, *args, input=None, **kwargs):
            self._flags = {'recursive_get': 1, 'components_initialized': 0}
            module_class.__init__(self, *args, **kwargs)
            PyTorch_Model.__init__(self, components, input=input, skip_module_init=True)
            PyTorch_Model._sync_parameters(self)

        def __call__(self, message, *args, **kwargs):
            try: # This will trigger a recursive call if possible.
                message = self.input(message, *args, **kwargs)
            except:
                pass

            return self.forward(message, *args, **kwargs)

    return ModelFromModule

def to_parameter(component):
    """
    Attempts to convert a component to Pytorch Parameter if it is a tensor-like. This is required for using that component during model training.
    """
    if not isinstance(component, Parameter) and not isinstance(component, Module): # Convert to Parameter
        try:
            component = Parameter(torch.Tensor(component))
        except:
        # If the component is not a tensor-like, Parameter, or Module, then it is some other object that we simply attach to the model
        # For example, it could be a Pipe or Junction that the model can call upon.
            pass
    return component

class PyTorch_Model(Module, Model, PyTorch_Junction ):

    def __init__(self, components={}, *args, input=None, skip_module_init=False, **kwargs):

        self._flags = {'recursive_get': 1, 'components_initialized': 0} # Used for controlling recursion. Don't mess with this.

        if not skip_module_init: # This is so the ModelFromModule Class can work.
            Module.__init__(self)

        self.components = PyTorch_Component_Map({}, model=self)
        self.init_default_components()

        for key, value in components.items():
            self.components[key] = value

        HookedPassThroughPipe.__init__(self, input = input)
        self.enable_updates()
        self.enable_inference()
        self._flags['components_initialized'] = 1

    def __setattr__(self, name, value):
        """
        Attribute modifications ignore the recursive aspect of Pipes.
        """

        if name in ['_flags', 'input']:
            object.__setattr__(self, name, value)
        else:
            if self._flags['components_initialized']:
                # self.update_components()
                self.components[name] = value
            self._flags['recursive_get'] = 0
            Module.__setattr__(self, name, value)
            self._flags['recursive_get'] = 1

    def __getattr__(self, name):

        return Model.__getattr__(self, name)

    def __call__(self, *args, **kwargs):

        return Model.__call__(self, *args, **kwargs)

    def state_dict(self):

        state_dict = {
            key: value for key, value in Module.state_dict(self).items()
            if key not in self.components._external_attribute_names.keys()
            }

        return state_dict

    def all_parameters(self):
        """
        Returns a list of every PyTorch parameter that this Model depends on that is unfrozen.
        This is useful for providing a parameters list to an optimizer.
        """

        all_params = []
        for param in self.components.values():
            if type(param) is Parameter:
                all_params.append(param)
            elif isinstance(param,Module):
                all_params.extend(list(param.parameters()))

        if hasattr(self, 'input') and self.input is not None:
            try:
                all_params.extend(self.input.all_parameters())
            except AttributeError:
                pass

        return all_params

    def _sync_parameters(self): # TODO: Test this.
        """
        Syncs the Parameters and Modules associated with this object with the component map.
        """
        for key, value in self._parameters.items():
            self.components[key] = value
        for key, value in self._modules.items():
            self.components[key] = value

    def set_state(self, state, reset=True):

        if reset:
            self.components = PyTorch_Component_Map({}, model=self)
            self.init_default_components()
        self.components.set_state(state)

    def get_state(self):
        """
        Returns state after serializing internal state.
        """
        state = self.components.get_state()
        serialized_state = {'external': state['external'], 'internal': {}}

        for key, value in state['internal'].items():
            if isinstance(value, Model): # Serialize the Model recursively and use that as the value.
                serialized_state['internal'][key] = value.get_state()
            elif isinstance(value, Module): # Serialize the submodule and use that as the value.
                state_dict = value.state_dict()
                for k, v in state_dict.items():
                    state_dict[k] = v.cpu().detach().numpy()
                serialized_state['internal'][key] = state_dict

            else:
                if type(value) in [torch.Tensor, torch.nn.Parameter]:
                    value = value.cpu().detach().numpy()
                serialized_state['internal'][key] = value

        return serialized_state

    def freeze(self, components = None):
        """
        Freezes the given components of the model (or all of them if none are specified) in order to prevent gradient updates.
        This means setting requires_grad to false for specified components so that these components
        are not updated during training.
        """
        self._change_temperature(False, components)

    def unfreeze(self, components = None):
        """
        Unfreezes the given components of the model (or all of them if none are specified) in order to prevent gradient updates.
        This means setting requires_grad to true for specified components so that these components
        are updated during training.
        """
        self._change_temperature(True, components)

    def _change_temperature(self, boo, components = None):

        if isinstance(components, str): # eg. a single component is provided as a string
            components = [components]

        if components is None:
            components = self.components

        for component in components:
            if isinstance(getattr(self, component), Parameter):
                getattr(self, component).requires_grad = boo
            elif isinstance(getattr(self,component), Model):
                getattr(self,component)._change_temperature(boo) # Recursively freezes Models
            elif isinstance(getattr(self,component), Module): # Is a PyTorch module but not a model.
                _change_temperature(boo, getattr(self,component))



from fireworks.core import Junction, Model, Message
from fireworks.utils import subset_dict
import torch
from torch import optim
from ignite.engine import Engine, Events
from ignite.handlers import ModelCheckpoint, Timer
from ignite.metrics import Loss
import visdom
import datetime
import numpy as np
import types
import logging
from copy import deepcopy

logging.getLogger(name=__name__)

def default_training_closure(model, optimizer, loss_fn):
    """
    This function produces a simple training loop that can be used for many situations. During each loop, the model is applied to the
    current batch, the loss_fn computes a loss, gradients are computed on the loss, and the optimizer updates its parameters using those
    gradients. The update function returns a dictionary containing the loss, the current value of the optimizer (ie. the gradients), and
    the output of the model.
    """
    def update_function(engine, batch):

        model.train()
        optimizer.zero_grad()
        output = model(batch)
        loss = loss_fn(output)
        loss.backward()
        optimizer.step()

        return {'loss': loss.detach().cpu().numpy(), 'optimizer': optimizer.state_dict(), 'state': model.get_state(), 'output': output}

    return update_function

def default_evaluation_closure(model, optimizer, loss_fn):
    """
    This function produces an evaluation loop that evaluates the model on the input data and computes the loss function but does not update
    the model. The optimizer is ignored.
    """
    def eval_function(engine, batch):
        output = model(batch)
        loss = loss_fn(output)

        return {'loss': loss.detach().cpu().numpy(), 'state': model.get_state(), 'output': output}

    return eval_function

class IgniteJunction(Junction):
    """
    This abstracts the functionality of an Ignite Engine into Junction format. See the Ignite documentation (https://github.com/pytorch/ignite)
    for more details.
    """
    required_components = {'model': Model, 'dataset': object}

    # These dictionaries describe the allowed optimizers and learning rate schedulers, along with the keyword arguments that each can accept.
    # These are all the optimizers/schedulers that PyTorch includes.

    optimizers = {
        'Adadelta':  optim.Adadelta,
        'Adagrad': optim.Adagrad,
        'Adam': optim.Adam,
        'SparseAdam': optim.SparseAdam,
        'Adamax': optim.Adamax,
        'ASGD': optim.ASGD,
        'LBFGS': optim.LBFGS,
        'RMSprop': optim.RMSprop,
        'Rprop': optim.Rprop,
        'SGD': optim.SGD,
    }
    optimizer_kwargs = {
        'Adadelta':  ['lr', 'rho', 'eps', 'weight_decay'],
        'Adagrad': ['lr', 'lr_decay', 'weight_decay', 'initial_accumulator_value'],
        'Adam': ['lr', 'betas', 'eps', 'weight_decay', 'amsgrad'],
        'SparseAdam': ['lr', 'betas', 'eps'],
        'Adamax': ['lr', 'betas', 'eps', 'weight_decay'],
        'ASGD': ['lr', 'lambd', 'alpha', 't0', 'weight_decay'],
        'LBFGS': ['lr', 'max_iter', 'max_eval', 'tolerance_grad', 'tolerance_change', 'history_size', 'line_search_fn'],
        'RMSprop': ['lr', 'alpha', 'eps', 'weight_decay', 'momentum', 'centered'],
        'Rprop': ['lr', 'etas', 'step_sizes'],
        'SGD': ['lr', 'momentum', 'dampening', 'weight_decay', 'nesterov'],
    }
    schedulers = {
        'LambdaLR': optim.lr_scheduler.LambdaLR,
        'StepLR': optim.lr_scheduler.StepLR,
        'MultiStepLR': optim.lr_scheduler.MultiStepLR,
        'ExponentialLR': optim.lr_scheduler.ExponentialLR,
        'CosineAnnealingLR': optim.lr_scheduler.CosineAnnealingLR,
        'ReduceLROnPlateau': optim.lr_scheduler.ReduceLROnPlateau,
    }
    scheduler_kwargs = {
        'LambdaLR': ['lr_lambda', 'last_epoch'],
        'StepLR': ['step_size', 'gamma', 'last_epoch'],
        'MultiStepLR': ['milestones', 'gamma', 'last_epoch'],
        'ExponentialLR': ['gamma', 'last_epoch'],
        'CosineAnnealingLR': ['T_max', 'eta_min', 'last_epoch'],
        'ReduceLROnPlateau': ['mode', 'factor', 'patience', 'verbose', 'threshold', 'threshold_mode', 'cooldown', 'min_lr', 'eps'],
    }

    def __init__(self, components, loss, optimizer, scheduler=None, update_function=default_training_closure, visdom=True, environment='default', description='', **kwargs):

        Junction.__init__(self, components = components)
        parameters = [x for x in self.model.all_parameters() if x.requires_grad]
        # Initialize engine
        self.optimizer = self.optimizers[optimizer](parameters, **subset_dict(kwargs, self.optimizer_kwargs[optimizer]))
        if scheduler is not None:
            self.optimizer = self.schedulers[scheduler](self.optimizer, **subset_dict(kwargs, self.scheduler_kwargs[scheduler]))
        self.loss = loss
        self.update_function = update_function(self.model, self.optimizer, self.loss)
        self.engine = Engine(self.update_function)

        # Configure metrics and events
        if not visdom:
            environment = None
        self.attach_events(environment=environment, description=description)

    def train(self, dataset = None, max_epochs=10):

        dataset = dataset or self.dataset
        self.engine.run(dataset, max_epochs=max_epochs)

    def run(self, *args, **kwargs):
        """ Alias for train """
        self.train(*args, **kwargs)

    def add_event_handler(self, *args, **kwargs):

        self.engine.add_event_handler(*args, **kwargs)

    def has_event_handler(self, *args, **kwargs):

        return self.engine.has_event_handler(*args, **kwargs)

    def attach_events(self, description, environment=None, save_file = None):

        tim = Timer()
        tim.attach( self.engine,
                    start=Events.STARTED,
                    step=Events.ITERATION_COMPLETED,
        )

        log_interval = 100
        plot_interval = 10

        @self.engine.on(Events.ITERATION_COMPLETED)
        def log_training_loss(engine):
            iter = (engine.state.iteration -1)
            if iter % log_interval == 0:
                logging.info("Epoch[{}] Iteration: {} Time: {} Loss: {:.2f}".format(
                    engine.state.epoch, iter, str(datetime.timedelta(seconds=int(tim.value()))), engine.state.output['loss']
                ))
        if environment:

            vis = visdom.Visdom(env=environment)

            def create_plot_window(vis, xlabel, ylabel, title):
                return vis.line(X=np.array([1]), Y=np.array([np.nan]), opts=dict(xlabel=xlabel, ylabel=ylabel, title=title))

            train_loss_window = create_plot_window(vis, '#Iterations', 'Loss', 'Training Loss {0}'.format(description))

            @self.engine.on(Events.ITERATION_COMPLETED)
            def plot_training_loss(engine):
                iter = (engine.state.iteration -1)
                if iter % plot_interval == 0:
                    vis.line(X=np.array([engine.state.iteration]),
                            Y=np.array([engine.state.output['loss']]),
                            update='append',
                            win=train_loss_window)


        # @self.engine.on(Events.ITERATION_COMPLETED)
        # def log_model_state(engine):
        #     iter = (engine.state.iteration-1)
        #     if iter % log_interval == 0:
        #         current_state = Message.from_objects(deepcopy(engine.state.output['state']))
        #         current_state['iteration'] = [iter]
        #         self.model_state = self.model_state.append(current_state)

        # if save_file is not None:
        #     save_interval = 50
        #     handler = ModelCheckpoint('/tmp/models', save_file, save_interval = save_interval, n_saved=5, create_dir=True, require_empty=False)
        #     self.engine.add_event_handler(Events.ITERATION_COMPLETED, handler, {'model': model})



import fireworks
import os
import pandas as pd
from fireworks.core.pipe import recursive
from fireworks import Message, Junction, Pipe, Model, PyTorch_Model, model_from_module
from fireworks.utils import index_to_list
import numpy as np
import math
import itertools
from fireworks.utils.exceptions import ParameterizationError
from fireworks.toolbox.pipes import BatchingPipe, LoopingPipe, ShufflerPipe, RepeaterPipe
from fireworks import Message, Junction
import random
import torch
from torch.nn import Parameter
from random import randint

test_dir = fireworks.test_dir
"""
This file contains numerous mock objects which are used by the tests.
"""

class one_way_dummy(Pipe):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.count = 0

    def __next__(self):
        self.count += 1
        if self.count <= 20:
            return {'count': np.array([self.count-1])}
        else:
            raise StopIteration# This will trigger StopIteration

    def reset(self):
        self.count = 0

class one_way_iter_dummy(one_way_dummy):

    def __iter__(self):
        self.reset()
        return self

class reset_dummy(Pipe):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.count = 0

    def reset(self):
        self.count = 0

class next_dummy(Pipe):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.count = 0

    def __next__(self):
        self.count += 1
        if self.count < 20:
            return {'count': np.array([self.count])}
        else:
            raise StopIteration # This will trigger StopIteration

class getitem_dummy(Pipe):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.length = 20

    def __getitem__(self, index):

        index = index_to_list(index)

        # if index == []:
        if len(index) == 0:
            return None
        elif max(index) < self.length and min(index) >= 0:
            return {'values': np.array(index)}
        else:
            raise IndexError("Out of bounds for dummy pipe with length {0}.".format(self.length))

    def __len__(self):
        return self.length

class recursion_dummy(Pipe):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.height = 0
        self.depth = 0

    @recursive()
    def jump(self, n=1):
        self.height += n
        return self.height

    @recursive(accumulate=True)
    def somersault(self, x):

        x = x or 0
        y = x+1
        self.depth += y
        return y

    def poop(self):
        return 3

class smart_dummy(Pipe):
    """
    Implements all of the methods in the above dummies
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.length = 20
        self.count = 0

    def __getitem__(self, index):

        index = index_to_list(index)

        if index == []:
            return None
        elif max(index) < self.length and min(index) >= 0:
            return Message({'values': np.array(index)})
        else:
            raise IndexError("Out of bounds for dummy pipe with length {0}.".format(self.length))

    def __call__(self, *args, **kwargs):

        return Message({'values': np.array([1,2,3])})

    def reset(self):
        self.count = 0
        return self

    def __next__(self):
        self.count += 1
        if self.count <= self.length:
            return Message({'values': np.array([self.count-1])})
        else:
            raise StopIteration# This will trigger StopIteration

    def __len__(self):
        return self.length

    def __iter__(self):
        return self.reset()

class ComplexModel(PyTorch_Model):

    required_components = ['activation']

    def __init__(self, components = {}, input = None, in_column = 'x', out_column = 'y', in_features=4, out_features=2):
        PyTorch_Model.__init__(self, components, input = input)
        self.in_column = in_column
        self.out_column = out_column
        self.in_features = in_features
        self.out_features = out_features
        self.components['layer1'] = torch.nn.modules.Linear(100, 23)

    def init_default_components(self):
        self.components['activation'] = torch.nn.Sigmoid()
        

    def forward(self, message):

        y = self.activation(self.linear(message[self.in_column]))
        message[self.out_column] = y

        return message 

class DummyModel(PyTorch_Model):
    """ Implements y = m*x + b """
    required_components = ['m', 'b']

    def __init__(self, components = {}, input = None, in_column = 'x', out_column = 'y'):
        PyTorch_Model.__init__(self, components, input = input)
        self.in_column = in_column
        self.out_column = out_column

    def init_default_components(self):
        """ Default y-intercept to 0 """
        self.components['b'] = Parameter(torch.Tensor([0.]))

    def forward(self, message):

        y = self.m*message[self.in_column]+self.b
        message[self.out_column] = y

        return message

class DummyMultilinearModel(PyTorch_Model):
    """ Implements y = m1*x1 +m2*x1 + b """
    required_components = ['m1', 'm2', 'b']

    def init_default_components(self):
        """ Default y-intercept to 0 """
        self.components['b'] = Parameter(torch.Tensor([0.]))

    def forward(self, message):

        y = self.m1*message['x1']+ self.m2*message['x2'] + self.b
        message['y'] = y
        return message

class LinearJunctionModel(PyTorch_Model):
    """ Implements y = f(x) + b, where the function f(x) is provided as a junction input. """

    required_components = ['b', 'f']

    def init_default_components(self):
        """ Default y-intercept to 0 """
        self.components['b'] = Parameter(torch.Tensor([0.]))

    def forward(self, message):

        y = self.f(message)['z'] + self.b
        message['y'] = y
        return message

class LinearModule(torch.nn.Module):
    """ Dummy PyTorch Module. """
    def __init__(self):
        torch.nn.Module.__init__(self)
        self.m = Parameter(torch.randn(1))
        self.b = Parameter(torch.randn(1))
        self.conv1 = torch.nn.Conv2d(1, 20, 5)

    def forward(self, message):

        message['y'] = self.m*message['x']+self.b

        return message

class DummyUpdateModel(PyTorch_Model):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self._count = 0

    def update(self, batch, **kwargs):

        self._count += len(batch)

    def forward(self, batch):
        return batch.append(batch)

class RandomJunction(Junction):

    def __call__(self, *args, **kwargs):

        target = random.sample(self.components.keys(),1)[0]
        return self.components[target](*args, **kwargs)

def generate_linear_model_data(n=300):
    """
    Generates n samples from a linear model with a small variability.
    """
    m = randint(-3,3)
    b = randint(-10,10)
    x = np.random.rand(n)*100
    errors = np.random.normal(0, .4, n) # Gaussian samples for errors
    y = m*x+b + errors

    return Message({'x':x, 'y_true':y}), {'m': m, 'b': b, 'errors': errors} # Second dict is for debugging

def generate_multilinear_model_data(n=550):
    """
    Generates n samples from a multilinear model y = m1*x1 + m2*x2 +b with a small variability.
    """

    m1 = randint(-5,5)
    m2 = randint(-5,5)
    b = randint(-10,10)
    x1 = np.random.rand(n)*100
    x2 = np.random.rand(n)*100
    errors = np.random.normal(0,.1,n) # Gaussian samples for errors
    y = m1*x1 + m2*x2 + b + errors

    return Message({'x1':x1, 'x2': x2, 'y_true':y}), {'m1': m1, 'm2':m2, 'b': b, 'errors': errors} # Second dict is for debugging



import numpy as np
from fireworks import Message
from fireworks.toolbox import preprocessing as pr
from fireworks.toolbox import CachingPipe, TensorPipe
from fireworks.utils.test_helpers import generate_linear_model_data
import numpy as np
from .pipes import ShufflerPipe, BatchingPipe
import math
from itertools import count
import torch 

def test_Normalizer():

    a = 10
    b = 3
    c = 20
    d = 4
    n = 1000
    data = Message({'ok': np.random.normal(a,b,n), 'good': np.random.normal(c,d,n)})
    shuffler = ShufflerPipe(data)
    batcher = TensorPipe(BatchingPipe(shuffler))
    normie = pr.Normalizer(input=batcher)
    normie.disable_inference()
    assert normie.count == 0
    if torch.cuda.is_available():
        normie.cuda()
    for batch in normie: # Loop through the dataset
        continue
    assert normie.count == n
    # assert normie.mean == {}
    # assert normie.variance == {}
    normie.compile()
    means = normie.mean
    variances = normie.variance
    assert (abs(means['ok'] - a) < .5).all()
    assert (abs(means['good'] - c) < .5).all()
    assert (abs(variances['ok'] - b**2) < 3).all()
    assert (abs(variances['good'] - d**2) < 6).all()
    normie.disable_updates()
    normie.enable_inference()
    batch = normie[0:100]
    batch = batch.to_dataframe()
    mok = np.mean(batch['ok'])
    vok = np.var(batch['ok'])
    mood = np.mean(batch['good'])
    vood = np.var(batch['good'])
    assert abs(mok )< .3
    assert abs(vok - 1 )< .4
    assert abs(mood )< .3
    assert abs(vok -1 )< .4
    normie2 = pr.Normalizer(input=normie)
    if torch.cuda.is_available():
        normie2.cuda()
    normie2.disable_inference()
    normie.enable_inference()
    normie.disable_updates()
    le = 0
    for batch in normie2:
        le += len(batch)
        continue
    normie2.compile()
    means = normie2.mean
    variances = normie2.variance
    assert (abs(means['ok']) < .5).all()
    assert (abs(means['good']) < .5).all()
    assert (abs(variances['ok'] - 1) < .5).all()
    assert (abs(variances['good'] - 1) < .5).all()

    assert normie.mean['good'] != normie2.mean['good']
    assert normie.mean['ok'] != normie2.mean['ok']
    state = normie.get_state()
    normie2.set_state(state, reset=False)
    assert normie.mean['good'] == normie2.mean['good']
    assert normie.mean['ok'] == normie2.mean['ok']

def test_train_test_split():

    # Test using a Message as input
    data, metadata = generate_linear_model_data(50)
    train, test = pr.train_test_split(data)
    assert len(train) == 40
    assert len(test) == 10
    train[0:20]
    test[0:4]
    for train_row,i in zip(train,count()): # Check that training set and test set are different
        for test_row,j in zip(test, count()):
             assert (train_row != test_row)

    # Test using a Pipe as input
    cache = CachingPipe(data)
    train2, test2 = pr.train_test_split(cache)
    train2[0:20]
    test2[0:3]

    for train_row in train2: # Check that training set and test set are different
        for test_row in test2:
            assert train_row != test_row



from functools import lru_cache
import numpy as np

@lru_cache(maxsize=32)
def one_hot(index, max):
    """
    This converts an integer to a one-hot vector, which is a format that is often used by statistical classifiers to represent predictions.
    Args:
        - index: The index in the one-hot array to set as 1
        - max: The size of the one-hot array.

    Returns:
        - hot: A one-hot array, which consists of all 0s except a 1 at a certain index which corresponds to the label classification.
    """
    hot = np.zeros(max)
    hot[index] = 1
    return hot

def index_to_list(index):
    """
    Converts an index to a list. This is used by some of the methods in message.py.
    """
    if type(index) is slice:
        index = slice_to_list(index)
    if type(index) is int:
        index = [index]
    return index

def slice_to_list(s):
    """
    Converts a slice object to a list of indices. This is used by some of the methods in message.py.
    """
    step = s.step or 1
    start = s.start
    stop = s.stop
    return [x for x in range(start,stop,step)]

def get_indices(values, listlike):
    """
    Returns the indices in litlike that match elements in values. This is used by some of the methods in message.py.
    """

    return [i for i, l  in zip(count(), listlike) if l in values]

def slice_length(orange):
    """
    Returns the length of the index corresponding to a slice.
    For example, slice(0,4,2) has a length of two.
    This is used by some of the methods in message.py.
    """
    t = type(orange)
    if t is slice:
        if orange.step:
            return int((orange.stop-orange.start)/orange.step) # This will also work for negative steps
        else: # Step size is 1 by default
            return orange.stop - orange.start
    else:
        return len(orange)

def subset_dict(dictionary, keys):
    """
    Returns a dict that contains all key,value pairs in dictionary where the key is one of the provided keys.
    This is used by some of the methods in message.py.    
    """
    keys = set(keys)
    return {key: value for key, value in dictionary.items() if key in keys}



from functools import lru_cache
import math
import numpy as np
from fireworks.toolbox import pipes as pl
from fireworks.core import Model, PyTorch_Model
from collections import defaultdict
import torch 
import random

"""
This file contains models that can perform common preprocessing tasks, such as batch normalization.
"""

def train_test_split(pipe, test=.2):
    """
    Splits input pipe into a training pipe and a test pipe. The indices representing the input pipe are shuffled, and assigned to the training
    and test sets randomly based on the proportions specified.

    Args:
        - pipe: A pipe which represents the data to be split up.
        - test: The proportion of the set that should be returns as test set. This should be between 0 and 1.

    Returns:
        - train_pipe: A pipe that represents the training data. You can call __getitem__, __next__, etc. on this pipe and it will transparently
                      provide elements from the shuffled training set.
        - test_pipe: Analogous to the train_pipe, this represents the test data, which is shuffled and disjoint from the training data.
    """
    if not hasattr(pipe, '__getitem__'):
        raise ValueError("Input pipe must be indexable via __getitem__")

    l = len(pipe)
    num_test = math.floor(l*test)
    indices = random.sample(range(l), l)    
    train_indices = indices[0:l-num_test]
    test_indices = indices[l-num_test:]

    test_pipe = pl.IndexMapperPipe(input=pipe, input_indices=range(0,len(test_indices)), output_indices=test_indices)
    train_pipe = pl.IndexMapperPipe(input=pipe, input_indices=range(0,len(train_indices)), output_indices=train_indices)

    return train_pipe, test_pipe

def oversample(): pass

def apply_noise(): pass

#IDEA: Instead of implementing this, what if we couple a LoopingSource + CachingSource to a SKLearn estimator?
# Doing so would create a path to seamless insertion of SKlearn modules into pipelines.
class Normalizer(PyTorch_Model):
    """
    Normalizes Data by Mean and Variance. Analogous to sklearn.preprocessing.Normalizer
    This Model uses a one-pass method to estimate the sample variance which is not guaranteed to be numerically stable.

    The functionality is implemented using hooks. Every time data is accessed from upstream pipes, this Model updates its estimate of the
    population mean and variance using the update() method. If self._inference_enabled is set to True, then the data will also be normalized
    based on those estimates. Means and variances are calculated on a per-column basis. You can also disable/enable the updating of these
    estimate by calling self.enable_updates / self.disable_updates.
    """

    required_components = ['mean', 'variance', 'count', 'rolling_sum', 'rolling_squares']

    def __init__(self, *args, **kwargs):

        PyTorch_Model.__init__(self, *args, **kwargs)
        self.freeze(['mean', 'variance', 'count', 'rolling_sum', 'rolling_squares'])

    def init_default_components(self):

        self.components['mean'] = defaultdict(lambda : [0])
        self.components['variance'] = defaultdict(lambda : 1)
        self.components['count'] = [0.]
        self.components['rolling_sum'] = defaultdict(lambda : 0.)
        self.components['rolling_squares'] = defaultdict(lambda : 0.)

    def forward(self, batch):
        """
        Uses computed means and variances in order to transform the given batch.
        """

        keys = self.mean.keys()
        for key in keys:
            if key in batch:
                if self.variance[key].device.type == 'cuda':
                    batch[key] = (batch[key] - self.mean[key])/torch.sqrt(self.variance[key])
                else:
                    batch[key] = (batch[key] - self.mean[key])/np.sqrt(self.variance[key])

        return batch

    def update(self, batch, method=None):
        """
        Updates internal tracking of mean and variance given a batch.
        """
        if method == 'next' or method is None:
            self.count += len(batch)
            for key in batch.keys(): #WARNING: This is numerically unstable
                self.rolling_sum[key] += sum(batch[key])
                self.rolling_squares[key] += sum((batch[key]-self.mean[key][0])**2)
                # self.rolling_squares[key] += np.var(batch[key])
                self.compile()

    def compile(self):
        """
        Computes mean and variance given internal rolling sum and squares.
        """
        for key in self.rolling_sum:
            self.mean[key] = self.rolling_sum[key] / self.count
            self.variance[key] = (self.rolling_squares[key] - (self.rolling_sum[key])**2/self.count) / self.count + self.mean[key]**2
            self.variance[key] = self.rolling_squares[key] / self.count

    # def fit(self, dataset=None, continuamos=False):
    #
    #     if dataset is None:
    #         dataset = self.input
    #
    #     if not continuamos:
    #         self.reset()
    #
    #     for batch in dataset:

    # def reset(self):
    #
    #     self.count = 0
    #     self.rolling_sum = defaultdict(lambda : 0)
    #     self.rolling_squares = defaultdict(lambda : 0)
    #
    #     try:
    #         self.recursive_call('reset')()
    #     except:
    #         pass



from ignite.handlers import ModelCheckpoint, Timer
import logging
import visdom

logging.getLogger(name=__name__)

"""
This code is largely adapted from the demos from the PyTorch and Ignite documentation.
See https://github.com/pytorch/ignite for more information.
"""

def visdom_loss_handler(modules_dict, model_name):
    """
    Attaches plots and metrics to trainer.
    This handler creates or connects to an environment on a running Visdom dashboard and creates a line plot that tracks the loss function of a
    training loop as a function of the number of iterations. This can be attached to an Ignite Engine, and the training closure must
    have 'loss' as one of the keys in its return dict for this plot to be made.
    See documentation for Ignite (https://github.com/pytorch/ignite) and Visdom (https://github.com/facebookresearch/visdom) for more information.
    """

    tim = Timer()
    tim.attach( trainer,
                start=Events.STARTED,
                step=Events.ITERATION_COMPLETED,
    )

    vis = visdom.Visdom(env=environment)
    def create_plot_window(vis, xlabel, ylabel, title):
        return vis.line(X=np.array([1]), Y=np.array([np.nan]), opts=dict(xlabel=xlabel, ylabel=ylabel, title=title))

    train_loss_window = create_plot_window(vis, '#Iterations', 'Loss', description)
    log_interval = 10

    @trainer.on(Events.ITERATION_COMPLETED)
    def log_training_loss(engine):
        iter = (engine.state.iteration -1)
        if iter % log_interval == 0:
            logging.info("Epoch[{}] Iteration: {} Time: {} Loss: {:.2f}".format(
                engine.state.epoch, iter, str(datetime.timedelta(seconds=int(tim.value()))), engine.state.output
            ))
        vis.line(X=np.array([engine.state.iteration]),
                 Y=np.array([engine.state.output]),
                 update='append',
                 win=train_loss_window)

    save_interval = 50
    handler = ModelCheckpoint('/tmp/models', model_name, save_interval = save_interval, n_saved=5, create_dir=True, require_empty=False)
    trainer.add_event_handler(Events.ITERATION_COMPLETED, handler, modules_dict)



from fireworks.extensions import training
from fireworks.toolbox.pipes import BatchingPipe, ShufflerPipe, FunctionPipe
from fireworks.utils.test_helpers import DummyModel, generate_linear_model_data
import torch

base_loss = torch.nn.MSELoss()
loss = lambda batch: base_loss(batch['y'], batch['y_true'])

def test_IgniteJunction():

    # Instantiate model
    model = DummyModel({'m': [1.]})
    # Instantiate dataste
    data, labels = generate_linear_model_data()
    to_tensor = lambda m: m.to_tensors()
    batcher = \
        FunctionPipe(
            BatchingPipe(
                ShufflerPipe(data),
            batch_size=50),
        function=to_tensor)
    # Instantiate engine
    junkie = training.IgniteJunction({'model': model, 'dataset': batcher}, loss=loss, optimizer='Adam')
    # Train and test that changes took place
    assert (model.m == 1.).all()
    assert (model.b == 0.).all()
    junkie.train()
    assert not (model.m == 1.).all()
    assert not (model.b == 0.).all()



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='fireworks-ml',
    version='0.3.3',
    packages=find_packages(),
    author_email="skhan8@mail.einstein.yu.edu",
    description="A batch-processing framework for data analysis and machine learning using PyTorch.",
    long_description=open('README.md').read(),
    url="https://github.com/smk508/fireworks",
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import os
from sqlalchemy import create_engine, Column, Float, Integer, String, DateTime
from sqlalchemy.orm import sessionmaker
import datetime
from fireworks.core import Message
from . import database as db
from deprecated import deprecated

"""
This module contains classes and functions for saving and loading data collected during experiments.
"""

metadata_columns = [
    Column('name', String),
    Column('iteration', Integer),
    Column('description', String),
    Column('timestamp', DateTime),
]
metadata_table = db.create_table('metadata', columns=metadata_columns)

def load_experiment(experiment_path): # TODO: clean up attribute assignments for loading
    """
    Returns an experiment object corresponding to the database in the given path.

    Args:
        experiment_path (str): Path to the experiment folder.

    Returns:
        experiment (Experiment): An Experiment object loaded using the files in the given folder path.
    """
    experiment_name = experiment_path.split('/')[-1]
    db_path = '/'.join(experiment_path.split('/')[:-1])
    return Experiment(experiment_name, db_path, load=True)

class Experiment:
    # NOTE: For now, we assume that the underlying database is sqlite on local disk
    # TODO: Expand to support nonlocal databases

    def __init__(self, experiment_name, db_path=".", description=None, load=False):

        self.name = experiment_name
        self.db_path = db_path
        self.description = description or ''
        self.timestamp = datetime.datetime.now() # QUESTION: Should this be updated on each load?
        self.engines = {}
        if load:
            self.load_experiment()
            self.engines = {
                name.rstrip('.sqlite'): self.get_engine(name.rstrip('.sqlite'))
                for name in os.listdir(self.save_path) if name.endswith('.sqlite')
                }
            self.load_metadata()
        else:
            self.create_dir()
            self.init_metadata()

        self.filenames = os.listdir(os.path.join(self.db_path,self.save_path)) # Refresh list of filenames

    def load_experiment(self, path=None, experiment_name=None):
        """
        Loads in parameters associated with this experiment from a directory.

        Args:
            path (str): Path to the experiment folder.
            experiment_name (str): Name to set this experiment to.

        """
        path = path or self.db_path
        experiment_name = experiment_name or self.name
        self.save_path = os.path.join(path, experiment_name)
        if not experiment_name in os.listdir(path):
            raise ValueError("Directory {exp_dir} was not found in {path}".format(exp_dir=experiment_name, path=path))
        self.engine = create_engine("sqlite:///{save_path}".format(save_path=os.path.join(self.save_path,'metadata.sqlite')))

    def create_dir(self):
        """
        Creates a folder in db_path directory corresponding to this Experiment.
        """
        dirs = os.listdir(self.db_path)
        previous_experiments = [d for d in dirs if d.startswith(self.name)]
        self.iteration = len(previous_experiments)
        os.makedirs(os.path.join(self.db_path, "{name}_{iteration}".format(name=self.name, iteration=self.iteration))) # TODO: Upgrade to 3.6 and use f-strings
        self.save_path = "{name}_{iteration}".format(name=self.name, iteration=self.iteration)
        self.engine = create_engine("sqlite:///{save_path}".format(save_path=os.path.join(self.db_path,self.save_path,'metadata.sqlite')))

    def load_metadata(self):
        """
        Loads metadata from experiment folder by reading the metadata table.
        """
        self.metadata = db.TablePipe(metadata_table, self.engine, columns=['name', 'iteration', 'description', 'timestamp'])
        # Session = sessionmaker(bind=self.engine)
        # session = Session()
        # assert False
        metadata = self.metadata.session.query(metadata_table).one()
        self.name = metadata.name
        self.iteration = metadata.iteration
        self.description = metadata.description
        self.timestamp = metadata.timestamp

    def init_metadata(self):
        """
        Initializes metadata table. This is a necessary action whenever using an SQLalchemy table for the first time and is idempotent,
        so calling this method multiple times does not produce side-effects.
        """
        self.metadata = db.TablePipe(metadata_table, self.engine, columns=['name', 'iteration', 'description', 'timestamp'])
        self.metadata.insert(Message({'name': [self.name], 'iteration': [self.iteration], 'description': [self.description], 'timestamp': [self.timestamp]}))
        self.metadata.commit()

    def get_engine(self, name):
        """
        Creates an engine corresponding to a database with the given name. In particular, this creates a file called {name}.sqlite
        in this experiment's save directory, and makes an engine to connect to it.

        Args:
            name: Name of engine to create. This will also be the name of the file that is created.

        Returns:
            engine: The new engine. You can also reach this engine now by calling self.engines[name]
        """
        self.engines[name] = create_engine("sqlite:///{filename}".format(filename=os.path.join(self.db_path,self.save_path, name+'.sqlite')))
        return self.engines[name]

    def get_session(self, name):
        """
        Creates an SQLalchemy session corresponding to the engine with the given name that can be used to interact with the database.

        Args:
            name: Name of engine corresponding to session. The engine will be created if one with that name does not already exist.

        Returns:
            session: A session created from the chosen engine.
        """
        if name in self.engines:
            engine = self.engines[name]
        else: # QUESTION: Should this raise an error or autocreate a new engine?
            engine = self.get_engine(name)
        Session = sessionmaker(bind=engine)
        session = Session()
        return session

    def open(self, filename, *args, string_only=False):
        """
        Returns a handle to a file with the given filename inside this experiment's directory.
        If string_only is true, then this instead returns a string with the path to create the file.
        If the a file with 'filename' is already present in the directory, this will raise an error.

        Args:
            filename (str): Name of file.
            args: Additional positional args for the open function.
            string_only (bool): If true, will return the path to the file rather than the file handle. This can be useful if you want to
                create the file using some other library.

        Returns:
            file: If string_only is True, the path to the file. Otherwise, the opened file handle. Note: You can use this method in a
                with statement to auto-close the file.
        """
        self.filenames = os.listdir(os.path.join(self.db_path,self.save_path)) # Refresh list of filenames
        path = os.path.join(self.db_path,self.save_path, filename)
        if string_only:
            return path
        else:
            return open(path, *args)
        self.filenames = os.listdir(os.path.join(self.db_path,self.save_path)) # Refresh list of filenames



from .pipe import Pipe, HookedPassThroughPipe
from .message import Message
from fireworks.utils import index_to_list
from fireworks.utils.test_helpers import *
import numpy as np
import itertools

test_dir = fireworks.test_dir

# def conforms_to_spec(pipe):
#
#     assert hasattr(pipe, '__iter__')
#     assert hasattr(pipe, '__next__')
#
#     return True
#
# def test_pipe(): pass

def test_PassThroughPipe():

    dumbo = smart_dummy()
    pishpish = Pipe(input=dumbo)

    assert pishpish.count == 0
    assert Message(pishpish.__next__()) == Message({'values': [0]})
    assert pishpish.count == 1
    pishpish.reset()
    assert pishpish.count == 0
    assert Message(pishpish[12]) == Message({'values': [12]})
    assert Message(pishpish[10:14]) == Message({'values': [10,11,12,13]})
    for i, j in zip(pishpish, itertools.count()):
        assert Message(i) == Message({'values': [j]})

    assert Message(i) == Message({'values': [19]})

    # Test private attributes
    dumbo._hii = 2
    dumbo.bye = 3
    assert hasattr(dumbo, '_hii')
    assert hasattr(dumbo, 'bye')
    assert not hasattr(pishpish, '_hii')
    assert hasattr(pishpish, 'bye')

def test_HookedPassThroughPipe():

    dumbo = smart_dummy()
    class Hooker(HookedPassThroughPipe):

        def _getitem_hook(self, message):

            message['interception'] = ['aho' for _ in range(len(message))]
            message.df = message.df.reindex(sorted(message.df.columns), axis=1)
            return message

        def _next_hook(self, message):

            message['interception'] = ['yaro' for _ in range(len(message))]
            message.df = message.df.reindex(sorted(message.df.columns), axis=1)
            return message

    pishpish = Hooker(input=dumbo)
    assert pishpish.count == 0
    assert Message(pishpish.__next__()) == Message({'values': [0], 'interception': ['yaro']})
    assert pishpish.count == 1
    pishpish.reset()
    assert pishpish.count == 0
    assert Message(pishpish[12]) == Message({'values': [12], 'interception': ['aho']})
    assert Message(pishpish[10:14]) == Message({'values': [10,11,12,13], 'interception': ['aho','aho','aho','aho']})
    pishpish.reset()
    for i, j in zip(pishpish, itertools.count()):
        assert i == Message({'values': [j], 'interception': ['yaro']})

    assert i == Message({'values': [19], 'interception': ['yaro']})

def test_recursive_decorator():

    alto = recursion_dummy()
    balto = recursion_dummy(alto)
    calto = recursion_dummy(balto)

    assert alto.height == 0 and alto.depth == 0
    assert balto.height == 0  and balto.depth == 0
    assert calto.height == 0  and calto.depth == 0

    calto.jump()
    assert calto.height == 1
    assert balto.height == 1
    assert alto.height == 1

    balto.jump()
    assert calto.height == 1
    assert balto.height == 2
    assert alto.height == 2

    calto.somersault(1)
    assert alto.depth == 1
    assert balto.depth == 2
    assert calto.depth == 3

    balto.somersault(1)
    assert alto.depth == 2
    assert balto.depth == 4

def test_setstate_getstate():

    dumbo = smart_dummy()
    dumbo.stateful_attributes = ['length', 'count']
    for i in range(10):
        next(dumbo)
    assert dumbo.count == 10
    assert dumbo.length == 20
    state = dumbo.get_state()
    assert state['internal']['count'] == 10
    assert state['internal']['length'] == 20
    new_state = {'internal': {'count': 15, 'length': 30}}
    dumbo.set_state(new_state)
    assert dumbo.count == 15
    assert dumbo.length == 30
    i = 15
    while True:
        try:
            dumbo.__next__()
            i += 1
        except StopIteration:
            break
    assert dumbo.count == 31
    assert i == 30



import fireworks
import os
from fireworks.toolbox import junctions as jn
from fireworks import Message
from fireworks.utils import index_to_list
from fireworks.utils.test_helpers import *
import numpy as np
import itertools

def test_junction():

    dumbo = one_way_dummy()
    bumbo = one_way_dummy()
    gumbo = one_way_dummy()

    angry = jn.RandomHubJunction(components={'dumbo': dumbo, 'bumbo': bumbo, 'gumbo': gumbo})
    angry.reset()
    assert angry._available_inputs == set(['dumbo', 'bumbo', 'gumbo'])
    numbaz = Message()

    counter = lambda l,i: sum([1 for x in l if x == i]) # Counts how often i appears in l

    while True:
        try:
            numbaz = numbaz.append(angry.__next__())
        except StopIteration:
            break
    assert dumbo.count == 21
    assert bumbo.count == 21
    assert gumbo.count == 21
    assert len(numbaz) == 60

    counts = {i:counter(numbaz['count'],i) for i in range(20)}

    for count in counts.values():
        assert count == 3 # Make sure each element showed up 3 times, corresponding to the 3 inputs

    mangry = jn.ClockworkHubJunction(components = {'dumbo': dumbo, 'bumbo': bumbo, 'gumbo': gumbo})
    bumbaz = Message()

    for nextone in mangry:
        bumbaz = bumbaz.append(nextone)

    assert dumbo.count == 21
    assert bumbo.count == 21
    assert gumbo.count == 21
    assert len(bumbaz) == 60

    for x,i in zip(bumbaz, itertools.count()):
        assert x['count'][0] == math.floor(i/3)

    counts = {i:counter(bumbaz['count'],i) for i in range(20)}

    for count in counts.values():
        assert count == 3 # Make sure each element showed up 3 times, corresponding to the 3 inputs

def test_RandomHubJunction():

    a = Message({'x':np.random.rand(100)})
    b = Message({'x':np.random.rand(100)})
    rob = jn.RandomHubJunction(components={'a':a, 'b':b})
    i = 0
    for x in rob:
        i+=1
    assert i == 200
    for x in rob:
        i+=1
    assert i == 400



from fireworks import cache
from fireworks import Message, TensorMessage
import torch
import numpy as np
import pandas as pd

tensors = {
    'a': torch.Tensor([1,2,3]),
    'b': torch.Tensor([4,5,6]),
    }
vectors = {
    'c': np.array([7,8,9]),
    'd': np.array([10,11,12]),
}
tensors2 = {
    'a': torch.Tensor([13,14,15]),
    'b': torch.Tensor([16,17,18]),
    }
vectors2 = {
    'c': np.array([19,20,21]),
    'd': np.array([22,23,24]),
}

def test_pointer_adjustment():

    r = 5
    f = cache.pointer_adjustment_function(r)
    assert f(4) == 0
    assert f(5) == 1
    assert f(6) == 1
    r = [2,4,6]
    f = cache.pointer_adjustment_function(r)
    assert f(0) == 0
    assert f(2) == 1
    assert f(5) == 2
    assert f(6) == 3
    r = slice(3,6)
    f = cache.pointer_adjustment_function(r)
    assert f(0) == 0
    assert f(3) == 1
    assert f(4) == 2
    assert f(5) == 3

def test_slice_to_list():
    a = slice(0,10)
    l = cache.slice_to_list(a)
    assert l == [0,1,2,3,4,5,6,7,8,9]
    a = slice(0,10,2)
    l = cache.slice_to_list(a)
    assert l == [0,2,4,6,8]

def test_init_set():
    m = cache.UnlimitedCache()
    dummy_message = Message(tensors, vectors)
    assert len(m) == 0
    m[0] = dummy_message[0]
    assert len(m) == 1
    m[3:6] = dummy_message
    assert len(m) == 4
    assert m[0] == dummy_message[0]
    assert m[3:6] == dummy_message
    dummy_2 = Message(tensors2, vectors2)
    m[3:6] = dummy_2
    assert len(m) == 4
    assert m[0] == dummy_message[0]
    assert m[3:6] == dummy_2
    m[7:10] = dummy_message
    assert m[7:10] == dummy_message
    
def test_del():
    m = cache.UnlimitedCache()
    dummy_message = Message(tensors, vectors)
    m[0] = dummy_message[0]
    m[3:6] = dummy_message
    assert 0 in m
    assert 3 in m
    assert 4 in m
    assert 5 in m
    assert m[0] == dummy_message[0]
    # del m[0]
    # assert not 0 in m
    del m[3:5]
    assert m[5] == dummy_message[2]
    assert not 3 in m
    assert not 4 in m
    m[3:6] = dummy_message
    # assert False
    pointers = m.pointers.copy()
    del m[5:8]
    assert not 5 in m
    assert not 6 in m
    assert m[3:5] == dummy_message[0:2]

def test_permute():
    m = cache.UnlimitedCache()
    dummy_message = Message(tensors, vectors)
    m[3:6] = dummy_message
    assert m.cache == dummy_message
    m._permute([2,1,0])
    assert m.cache != dummy_message
    m._permute([2,1,0])
    assert m.cache == dummy_message

def test_LRUCache():
    m = cache.LRUCache(10, buffer_size=2)
    dummy_message = Message(tensors, vectors)
    m[2:5] = dummy_message
    assert m[2:5] == dummy_message
    assert m.rank_dict.keys() == m.pointers.keys()
    m[7:10] = dummy_message
    assert m[2:5] == dummy_message
    assert m[7:10] == dummy_message
    assert m.rank_dict.keys() == m.pointers.keys()
    m[12:15] = dummy_message
    assert m[2:5] == dummy_message
    assert m[7:10] == dummy_message
    assert m[12:15] == dummy_message
    assert m.rank_dict.keys() == m.pointers.keys()
    # At this point, the least recently used elements are in the beginning
    m[15:18] = dummy_message
    assert m[15:18] == dummy_message
    assert len(m) == 10
    assert m.rank_dict.keys() == m.pointers.keys()
    assert not 2 in m
    assert not 3 in m
    assert set([4,7,8,9,12,13,14,15,16,17]) == set(m.rank_dict.keys())
    m[[4,8,9,13,14,15,16]] # Trigger __getitem__; Now 7, 12 and 17 should be queueud for deletion
    m[18:21] = dummy_message
    assert len(m) == 10
    for i in [7,12,17]:
        assert not i in m
    # Trigger a buffer clearance.
    m[29] = dummy_message[0]
    assert len(m) == 9

def test_LFUCache():
    m = cache.LFUCache(10, buffer_size=2)
    dummy_message = Message(tensors, vectors)
    m[2:5] = dummy_message
    assert m[2:5] == dummy_message
    assert m.rank_dict.keys() == m.pointers.keys()
    m[7:10] = dummy_message
    assert m[2:5] == dummy_message
    assert m[7:10] == dummy_message
    assert m.rank_dict.keys() == m.pointers.keys()
    m[12:15] = dummy_message
    assert m[2:5] == dummy_message
    assert m[7:10] == dummy_message
    assert m[12:15] == dummy_message
    assert m.rank_dict.keys() == m.pointers.keys()
    # At this point, 12:15 are the least frequently used elements.
    m[15:18] = dummy_message
    assert m[15:18] == dummy_message
    assert len(m) == 10
    assert m.rank_dict.keys() == m.pointers.keys()
    assert not 12 in m
    assert not 13 in m
    assert set([2,3, 4,7,8,9,14,15,16,17]) == set(m.rank_dict.keys())
    # Trigger __getitem__
    m[29] = dummy_message[0]
    assert len(m) == 9
